from typing import TypeVar, Any

from untypy.error import UntypyAttributeError
from untypy.impl.protocol import ProtocolChecker
from untypy.impl.wrappedclass import WrappedType
from untypy.interfaces import CreationContext


class BoundGenericProtocolChecker(ProtocolChecker):
    def protocol_type(self) -> str:
        return "bound generic"


def WrappedGenericAlias(alias, ctx: CreationContext):
    typevars = dict(zip(alias.__origin__.__parameters__, alias.__args__))
    for key, value in typevars.items():
        _checktypevar(key, value)
    ctx = ctx.with_typevars(typevars)
    # This WrappedType must also be a generic alias.
    # So it can be as Typennotation
    tname = []
    for t in typevars:
        a = ctx.find_checker(typevars[t])
        if a is None:
            tname.append(str(typevars[t]))
        else:
            tname.append(a.describe())

    wt = WrappedType(alias.__origin__, ctx, name=f"{alias.__origin__.__name__}[" + (', '.join(tname)) + "]")
    wt.__origin__ = alias.__origin__
    wt.__args__ = alias.__args__
    return wt


def _checktypevar(typevar: TypeVar, bound: Any) -> None:
    # See https://www.python.org/dev/peps/pep-0484/#covariance-and-contravariance
    if len(typevar.__constraints__) > 0:
        if typevar.__covariant__ == True:
            if not issubclass(bound, typevar.__constraints__):
                raise UntypyAttributeError(
                    f"Violation in TypeVar {typevar}: {bound} must be a covariant of one of type {typevar.__constraints__}."
                )
        elif typevar.__contravariant__ == True:
            found = False
            for c in typevar.__constraints__:
                if issubclass(c, bound):
                    found = True
            if not found:
                raise UntypyAttributeError(
                    f"Violation in TypeVar {typevar}: {bound} must be a contravariant of one of type {typevar.__constraints__}."
                )
        else:
            if not bound in typevar.__constraints__:
                raise UntypyAttributeError(
                    f"Violation in TypeVar {typevar}: {bound} must be a exactly of one of type {typevar.__constraints__}."
                    f"\nYou may want to use TypeVar(..., bound=...) instead."
                )
    elif typevar.__bound__ is not None:
        if not type(typevar.__bound__) is type:
            raise UntypyAttributeError(
                f"Bound {bound} of TypeVar {typevar} must be a type."
            )
        if not issubclass(bound, typevar.__bound__):
            raise UntypyAttributeError(
                f"Violation in TypeVar {typevar}: {bound} must be a subclass of {typevar.__bound__}."
            )



from . import writeYourProgram as w

# Exported names that are available for star imports (in alphabetic order)
Any = w.Any
Callable = w.Callable
Generator = w.Generator
Iterable = w.Iterable
Iterator = w.Iterator
Literal = w.Literal
Mapping = w.Mapping
Optional = w.Optional
Sequence = w.Sequence
Protocol = w.Protocol
Union = w.Union
check = w.check
dataclass = w.dataclass
floatNegative = w.floatNegative
floatNonNegative = w.floatNonNegative
floatNonPositive = w.floatNonPositive
floatPositive = w.floatPositive
intNegative = w.intNegative
intNonNegative = w.intNonNegative
intNonPositive = w.intNonPositive
intPositive = w.intPositive
math = w.math
nat = w.nat
record = w.record
T = w.T
U = w.U
unchecked = w.unchecked
V = w.V

__all__ = [
    'Any',
    'Callable',
    'Generator',
    'Iterable',
    'Iterator',
    'Literal',
    'Mapping',
    'Optional',
    'Sequence',
    'Protocol',
    'Union',
    'check',
    'dataclass',
    'floatNegative',
    'floatNonNegative',
    'floatNonPositive',
    'floatPositive',
    'intNegative',
    'intNonNegative',
    'intNonPositive',
    'intPositive',
    'math',
    'nat',
    'record',
    'T',
    'U',
    'unchecked',
    'V'
]

# Exported names not available for star imports (in alphabetic order)
Lock = w.Lock
LockFactory = w.LockFactory
initModule = w.initModule
printTestResults = w.printTestResults
resetTestCount = w.resetTestCount



import inspect
import types
from typing import Optional, Union, List

from untypy.error import UntypyTypeError, Frame, Location
from untypy.interfaces import ExecutionContext, TypeChecker, WrappedFunction
from untypy.util.display import IndicatorStr
from untypy.util.return_traces import get_last_return


class ReplaceTypeExecutionContext(ExecutionContext):

    def __init__(self, upper: Optional[ExecutionContext], name: str):
        self.upper = upper
        self.name = name

    def wrap(self, err: UntypyTypeError) -> UntypyTypeError:
        err = err.with_frame(Frame(
            self.name,
            None, None, None
        ))

        if self.upper is not None:
            err = self.upper.wrap(err)
        return err


class CompoundTypeExecutionContext(ExecutionContext):
    upper: ExecutionContext
    checkers: list[TypeChecker]
    idx: int

    def __init__(self, upper: ExecutionContext, checkers: list[TypeChecker], idx: int):
        self.upper = upper
        self.checkers = checkers
        self.idx = idx

    def declared(self) -> Optional[Location]:
        return None

    def responsable(self) -> Optional[Location]:
        return None

    def name(self) -> str:
        raise NotImplementedError

    def wrap(self, err: UntypyTypeError) -> UntypyTypeError:
        type_declared = self.name() + "["
        indicator = " " * len(type_declared)

        for i, checker in enumerate(self.checkers):
            if i == self.idx:
                next_type, next_indicator = err.next_type_and_indicator()
                type_declared += next_type
                indicator += next_indicator
            else:
                type_declared += checker.describe()
                indicator += " " * len(checker.describe())

            if i != len(self.checkers) - 1:  # not last element
                type_declared += ", "
                indicator += "  "

        type_declared += "]"

        err = err.with_frame(Frame(
            type_declared,
            indicator,
            declared=self.declared(),
            responsable=self.responsable(),
        ))

        return self.upper.wrap(err)


class NoResponsabilityWrapper(ExecutionContext):
    upper: ExecutionContext

    def __init__(self, upper: ExecutionContext):
        self.upper = upper

    def wrap(self, err: UntypyTypeError) -> UntypyTypeError:
        full = self.upper.wrap(err)

        # now remove responsability in frames:
        frames_to_add = []
        for frame in full.frames:
            if frame not in err.frames:
                frame.responsable = None
                frames_to_add.append(frame)

        for frame in frames_to_add:
            err = err.with_frame(frame)

        for note in full.notes:
            err = err.with_note(note)

        if full.previous_chain is not None:
            err = err.with_previous_chain(full.previous_chain)

        return err


class ReturnExecutionContext(ExecutionContext):
    fn: WrappedFunction

    def __init__(self, fn: WrappedFunction):
        self.reti_loc = get_last_return()
        self.fn = fn

    def wrap(self, err: UntypyTypeError) -> UntypyTypeError:
        (next_ty, indicator) = err.next_type_and_indicator()
        return_id = IndicatorStr(next_ty, indicator)

        original = WrappedFunction.find_original(self.fn)

        try:
            signature = inspect.signature(original)  # TODO:!!! FIX BUILTINS
            front_sig = []
            for name in signature.parameters:
                front_sig.append(f"{name}: {self.fn.checker_for(name).describe()}")
            front_sig = f"{format_name(original)}(" + (", ".join(front_sig)) + ") -> "
            return_id = IndicatorStr(front_sig) + return_id
        except:
            return_id = IndicatorStr("???")

        declared = WrappedFunction.find_location(self.fn)
        responsable = declared

        if responsable is not None:
            if err.expected is not None and err.given is None:
                # Missing Return-Value?
                err = err.with_note("Did you miss a return statement?")
                last_line = responsable.line_no + responsable.line_span - 1
                responsable = responsable.narrow_in_span((responsable.file, last_line))
            else:
                responsable = responsable.narrow_in_span(self.reti_loc)

        return err.with_frame(Frame(
            return_id.ty,
            return_id.indicator,
            declared=declared,
            responsable=responsable,
        ))

def format_name(orig):
    n = orig.__name__
    if inspect.isclass(orig):
        k = 'class'
        if hasattr(orig, '__kind'):
            k = getattr(orig, '__kind')
        if k:
            return f"{k} constructor {n}"
    return n


class ArgumentExecutionContext(ExecutionContext):
    n: WrappedFunction
    stack: inspect.FrameInfo
    argument_name: str

    def __init__(self,
                 fn: Union[WrappedFunction, types.FunctionType],
                 stack: Optional[inspect.FrameInfo],
                 argument_name: str,
                 declared: Optional[Location] = None,
                 upper: ExecutionContext = None):
        self.fn = fn
        self.stack = stack
        self.argument_name = argument_name
        self.declared = declared
        self.upper = upper

    def wrap(self, err: UntypyTypeError) -> UntypyTypeError:
        (next_ty, indicator) = err.next_type_and_indicator()
        error_id = IndicatorStr(next_ty, indicator)

        original = WrappedFunction.find_original(self.fn)
        try:
            signature = inspect.signature(original)
        except ValueError:
            # fails on some built-ins
            signature = inspect.signature(self.fn)

        wf = None
        if (hasattr(self.fn, '__wf')):
            wf = getattr(self.fn, '__wf')
        elif isinstance(self.fn, WrappedFunction):
            wf = self.fn

        arglist = []
        for name in signature.parameters:
            if name is self.argument_name:
                arglist.append(IndicatorStr(f"{name}: ") + error_id)
            else:
                if wf is not None:
                    arglist.append(IndicatorStr(f"{name}: {wf.checker_for(name).describe()}"))
                else:
                    arglist.append(IndicatorStr(f"{name}"))

        id = IndicatorStr(f"{format_name(original)}(") + IndicatorStr(", ").join(arglist)

        if wf is not None:
            id += IndicatorStr(f") -> {wf.checker_for('return').describe()}")
        else:
            id += IndicatorStr(f")")

        if self.declared is None:
            declared = WrappedFunction.find_location(self.fn)
        else:
            declared = self.declared

        if self.stack is not None:
            responsable = Location.from_stack(self.stack)
        else:
            responsable = None

        frame = Frame(
            id.ty,
            id.indicator,
            declared=declared,
            responsable=responsable
        )
        if self.upper:
            err = self.upper.wrap(err)
        return err.with_frame(frame)


class GenericExecutionContext(ExecutionContext):
    def __init__(self, *, declared: Union[None, Location, List[Location]] = None,
                 responsable: Union[None, Location, List[Location]] = None,
                 upper_ctx: Optional[ExecutionContext] = None):
        self.declared = declared
        self.responsable = responsable
        self.upper_ctx = upper_ctx

    def wrap(self, err: UntypyTypeError) -> UntypyTypeError:
        declared = []
        if isinstance(self.declared, Location):
            declared.append(self.declared)
        if isinstance(self.declared, list):
            declared.extend(self.declared)

        responsable = []
        if isinstance(self.responsable, Location):
            responsable.append(self.responsable)
        if isinstance(self.responsable, list):
            responsable.extend(self.responsable)

        while len(declared) < len(responsable): declared.append(None)
        while len(declared) > len(responsable): responsable.append(None)

        for (d, r) in zip(declared, responsable):
            (t, i) = err.next_type_and_indicator()
            err = err.with_frame(Frame(t, i, d, r))

        if self.upper_ctx is not None:
            return self.upper_ctx.wrap(err)
        else:
            return err



import inspect


class IndicatorStr:
    ty: str
    indicator: str

    def __init__(self, ty: str, indicator: str = ""):
        n = 0 if ty is None else len(ty)
        while len(indicator) < n:
            indicator += " "

        self.ty = ty
        self.indicator = indicator

    def __add__(self, other):
        return IndicatorStr(self.ty + other.ty, self.indicator + other.indicator)

    def join(self, lst):
        return IndicatorStr(
            self.ty.join(map(lambda s: s.ty, lst)),
            self.indicator.join(map(lambda s: s.indicator, lst)),
        )


def format_argument_values(args, kwargs):
    allargs = []
    for a in args:
        allargs.append(a.__repr__())
    for k,v in kwargs.items():
        allargs.append(k + "=" + v.__repr__())

    return "(" + ", ".join(allargs) + ")"


def withIndefiniteArticle(s):
    if s:
        first = s[0]
        if first in "aeiouyAEIOUY":
            return 'an ' + s
        else:
            return 'a ' + s
    else:
        return s



import inspect
import sys
from typing import Callable, Dict, Optional

from untypy.error import UntypyAttributeError, UntypyNameError, UntypyTypeError
from untypy.impl.any import SelfChecker
from untypy.interfaces import WrappedFunction, TypeChecker, CreationContext, WrappedFunctionContextProvider, \
    ExecutionContext
from untypy.util import ArgumentExecutionContext, ReturnExecutionContext
from untypy.util.typehints import get_type_hints


class TypedFunctionBuilder(WrappedFunction):
    inner: Callable
    signature: inspect.Signature
    _checkers: Optional[Dict[str, TypeChecker]]

    special_args = ['self', 'cls']
    method_name_ignore_return = ['__init__']

    def __init__(self, inner: Callable, ctx: CreationContext):
        self.inner = inner
        self.signature = inspect.signature(inner)
        self.ctx = ctx
        self.fc = None
        self._checkers = None

        try:
            # try to detect errors like missing arguments as early as possible.
            # but not all type annotations are resolvable yet.
            # so ignore `UntypyNameError`s
            self.checkers()
        except UntypyNameError:
            pass

        if hasattr(self.inner, "__fc"):
            self.fc = getattr(self.inner, "__fc")

    def checkers(self) -> Dict[str, TypeChecker]:
        if self._checkers is not None:
            return self._checkers

        annotations = get_type_hints(self.inner, self.ctx)

        checkers = {}
        checked_keys = list(self.signature.parameters)

        # Remove self and cls from checking
        if len(checked_keys) > 0 and checked_keys[0] in self.special_args:
            checkers[checked_keys[0]] = SelfChecker()
            checked_keys = checked_keys[1:]

        for key in checked_keys:
            if self.signature.parameters[key].annotation is inspect.Parameter.empty:
                raise self.ctx.wrap(
                    UntypyAttributeError(f"Missing annotation for argument '{key}' of function {self.inner.__name__}\n"
                                         "Partial annotation are not supported."))
            annotation = annotations[key]
            checker = self.ctx.find_checker(annotation)
            if checker is None:
                raise self.ctx.wrap(UntypyAttributeError(f"\n\tUnsupported type annotation: {annotation}\n"
                                                         f"\tin argument '{key}'"))
            else:
                checkers[key] = checker

        if self.inner.__name__ in self.method_name_ignore_return:
            checkers['return'] = SelfChecker()
        else:
            if not 'return' in annotations:
                annotation = None
            else:
                annotation = annotations['return']
            return_checker = self.ctx.find_checker(annotation)
            if return_checker is None:
                raise self.ctx.wrap(UntypyAttributeError(f"\n\tUnsupported type annotation: {annotation}\n"
                                                         f"\tin return"))

            checkers['return'] = return_checker

        self._checkers = checkers
        return checkers

    def build(self):
        def wrapper(*args, **kwargs):
            # first is this fn
            caller = sys._getframe(1)
            (args, kwargs, bindings) = self.wrap_arguments(lambda n: ArgumentExecutionContext(self, caller, n), args,
                                                           kwargs)
            ret = self.inner(*args, **kwargs)
            ret = self.wrap_return(ret, bindings, ReturnExecutionContext(self))
            return ret

        if inspect.iscoroutine(self.inner):
            raise UntypyAttributeError("Async functions are currently not supported.")
        else:
            w = wrapper

        setattr(w, '__wrapped__', self.inner)
        setattr(w, '__name__', self.inner.__name__)
        setattr(w, '__signature__', self.signature)
        setattr(w, '__wf', self)

        # Copy useful attributes
        # This is need for the detection of abstract classes
        for attr in ['__isabstractmethod__']:
            if hasattr(self.inner, attr):
                setattr(w, attr, getattr(self.inner, attr))

        return w

    def wrap_arguments(self, ctxprv: WrappedFunctionContextProvider, args, kwargs):
        try:
            bindings = self.signature.bind(*args, **kwargs)
        except TypeError as e:
            err = UntypyTypeError(header=str(e))
            raise ctxprv("").wrap(err)

        bindings.apply_defaults()
        if self.fc is not None:
            self.fc.prehook(bindings, ctxprv)
        for name in bindings.arguments:
            check = self.checkers()[name]
            ctx = ctxprv(name)
            bindings.arguments[name] = check.check_and_wrap(bindings.arguments[name], ctx)
        return bindings.args, bindings.kwargs, bindings

    def wrap_return(self, ret, bindings, ctx: ExecutionContext):
        check = self.checkers()['return']
        if self.fc is not None:
            self.fc.posthook(ret, bindings, ctx)
        return check.check_and_wrap(ret, ctx)

    def describe(self):
        return str(self.signature)

    def get_original(self):
        return self.inner

    def checker_for(self, name: str) -> TypeChecker:
        return self.checkers()[name]



from typing import Tuple


class DisplayMatrix:
    chars: dict

    def __init__(self, src: str, max_lines=5):
        self.chars = {}
        self.max_lines = max_lines
        for y, line in enumerate(src.splitlines()):
            for x, char in enumerate(line):
                self.chars[(x, y)] = char

    def write(self, pos: Tuple[int, int], message: str):
        for y, line in enumerate(message.splitlines()):
            for x, char in enumerate(line):
                self.chars[(x + pos[0], y + pos[1])] = char

    def __str__(self):
        # Slow, but this is only in the "error" path
        max_y = 0
        max_x = 0
        for x, y in self.chars:
            max_y = max(max_y, y)
            max_x = max(max_x, x)

        max_y = min(max_y, self.max_lines)
        out = ""
        for y in range(max_y + 1):
            for x in range(max_x + 1):
                if (x, y) in self.chars:
                    out += self.chars[(x, y)]
                else:
                    out += " "
            out += "\n"
        return out



# coding=utf-8
# NOTE: this file must have valid python 2 syntax. We want to display an error message
# when running with python 2.

import sys
pythonVersion = sys.version.split()[0]
if not pythonVersion.startswith('3.'):
    sys.stderr.write("\nERROR: Python 3 is required, not " + pythonVersion + ".\n\n")
    if sys.flags.interactive:
        import os
        os._exit(1)
    else:
        sys.exit(1)

if __name__ == '__main__':
    import runner
    runner.main(globals())



from typing import TypeVar, Generic, Iterable, Optional, Union, Any, Iterator

from untypy.impl.interfaces.util import overwrite
from untypy.interfaces import CreationContext
from untypy.util import ReturnExecutionContext

I = TypeVar("I")


def sig_getitem():
    def __getitem__(self, key: Union[int, slice]) -> Union[I, list[I]]:
        pass

    return __getitem__


def cast_wlist(lst) -> list:
    if hasattr(lst, '_WrappedClassFunction__inner'):
        return lst._WrappedClassFunction__inner
    else:
        return lst


class List(Generic[I], list):
    # doc @ https://docs.python.org/3/tutorial/datastructures.html
    # and https://docs.python.org/3/library/stdtypes.html#common-sequence-operations
    # Exact signatures are undocumented :/
    # HINT: Argument names must match.

    def append(self, object: I) -> None:
        pass

    def extend(self, iterable: Iterable[I]) -> None:
        pass

    def insert(self, i: int, x: I) -> None:
        pass

    def remove(self, x: I) -> None:
        pass

    def pop(self, i: int = -1) -> Optional[I]:
        pass

    def clear(self) -> None:
        pass

    def index(self, value: I, start: Optional[int] = 0, stop: Optional[int] = 9223372036854775807) -> int:
        # get index of list
        pass

    def count(self, value: I) -> int:
        pass

    def sort(self, *, key: Any = None, reverse: bool = False) -> None:
        # inner list will check type of key.
        pass

    def __contains__(self, key: I) -> bool:
        pass

    def __delitem__(self, i: Union[int, slice]):
        pass

    @overwrite("advanced")
    def __getitem__(self, ctx: CreationContext):
        # self is WrappedClassFunction

        u_checker = ctx.find_checker(Union[int, slice])
        inner_checker = ctx.find_checker(I)

        # TODO: implement some kind caching maybe.
        me_checker = lambda: ctx.find_checker(list[I])

        def inner(me, item):
            ret_ctx = me._WrappedClassFunction__return_ctx
            if ret_ctx is None:
                ret_ctx = ReturnExecutionContext(self)

            if isinstance(item, int):
                item = me._WrappedClassFunction__inner[item]
                return inner_checker.check_and_wrap(item, ret_ctx)
            elif isinstance(item, slice):
                item = me._WrappedClassFunction__inner[item]
                return item
            else:
                # TODO:
                raise NotImplementedError()

        setattr(self, '__original', sig_getitem())
        return inner

    @overwrite("simple")
    def __add__(self, other: Iterable) -> Any:
        return cast_wlist(self) + other

    @overwrite("simple")
    def __mul__(self, n: int) -> Any:
        return cast_wlist(self) * n

    def __iadd__(self, other: Iterable[I]) -> Any:  # returns self
        pass

    def __imul__(self, n: int) -> Any:  # returns self
        pass

    def __setitem__(self, key: Union[int, slice], value: Any) -> None:
        pass

    def __iter__(self) -> Iterator[I]:
        pass

    @overwrite("simple")
    def __radd__(self, other):
        return cast_wlist(other) + cast_wlist(self)

    @overwrite("simple")
    def __lt__(self, other):
        return cast_wlist(self).__lt__(cast_wlist(other))

    @overwrite("simple")
    def __le__(self, other):
        return cast_wlist(self).__le__(cast_wlist(other))

    @overwrite("simple")
    def __eq__(self, other):
        return cast_wlist(self).__eq__(cast_wlist(other))

    @overwrite("simple")
    def __ne__(self, other):
        return cast_wlist(self).__ne__(cast_wlist(other))

    @overwrite("simple")
    def __gt__(self, other):
        return cast_wlist(self).__gt__(cast_wlist(other))

    @overwrite("simple")
    def __ge__(self, other):
        return cast_wlist(self).__ge__(cast_wlist(other))

    # lower type on
    # __add__
    # __mul__
    # copy

    # Type fixed by std impl
    # __repr__
    # __reversed__
    # __len__
    # reverse



import inspect
import sys
from collections.abc import Callable as AbcCallable
from typing import Any, Optional, Callable, Union, Tuple

from untypy.error import UntypyTypeError, UntypyAttributeError, Frame, Location
from untypy.interfaces import TypeChecker, TypeCheckerFactory, CreationContext, ExecutionContext, WrappedFunction, \
    WrappedFunctionContextProvider
# These Types are prefixed with an underscore...
from untypy.util import ArgumentExecutionContext, ReturnExecutionContext

CallableTypeOne = type(Callable[[], None])
CallableTypeTwo = type(AbcCallable[[], None])


class CallableFactory(TypeCheckerFactory):

    def create_from(self, annotation: Any, ctx: CreationContext) -> Optional[TypeChecker]:
        if type(annotation) == CallableTypeOne or type(annotation) == CallableTypeTwo:
            return CallableChecker(annotation, ctx)
        else:
            return None


class CallableChecker(TypeChecker):
    return_checker: TypeChecker
    argument_checker: list[TypeChecker]

    def __init__(self, annotation: Union[CallableTypeOne, CallableTypeTwo], ctx: CreationContext):
        arguments_ty = annotation.__args__[:-1]
        return_ty = annotation.__args__[-1]

        return_checker = ctx.find_checker(return_ty)
        if return_checker is None:
            raise ctx.wrap(UntypyAttributeError(f"Return Type Annotation not found. {return_ty}"))

        argument_checker = []
        for arg in arguments_ty:
            checker = ctx.find_checker(arg)
            if checker is None:
                raise ctx.wrap(UntypyAttributeError(f"Argument Type Annotation not found. {arg}"))
            argument_checker.append(checker)

        self.return_checker = return_checker
        self.argument_checker = argument_checker

    def may_be_wrapped(self) -> bool:
        return True

    def base_type(self) -> list[Any]:
        return [Callable]

    def check_and_wrap(self, arg: Any, ctx: ExecutionContext) -> Any:
        if callable(arg):
            return TypedCallable(arg, self.return_checker, self.argument_checker, ctx)
        else:
            raise ctx.wrap(UntypyTypeError(arg, self.describe()))

    def describe(self) -> str:
        arguments = ", ".join(map(lambda e: e.describe(), self.argument_checker))
        return f"Callable[[{arguments}], {self.return_checker.describe()}]"


class TypedCallable(Callable, WrappedFunction):
    return_checker: TypeChecker
    argument_checker: list[TypeChecker]
    inner: Callable
    fn: Callable
    ctx: ExecutionContext

    def __init__(self, inner: Callable, return_checker: TypeChecker,
                 argument_checker: list[TypeChecker], ctx: ExecutionContext):
        # unwrap if wrapped function
        self.ResponsibilityType = None
        if hasattr(inner, '__wf'):
            inner = getattr(inner, '__wf')
        self.inner = inner
        self.return_checker = return_checker
        self.argument_checker = argument_checker
        self.ctx = ctx
        self.fn = WrappedFunction.find_original(self.inner)
        setattr(self, '__wf', self)

    def __call__(self, *args, **kwargs):
        caller = sys._getframe(1)

        (args, kwargs, bindings) = self.wrap_arguments(
            lambda i: TypedCallableArgumentExecutionContext(self, caller, i, self.ctx), args, kwargs)

        bind2 = None
        if isinstance(self.inner, WrappedFunction):
            (args, kwargs, bind2) = self.inner.wrap_arguments(lambda n:
                                                              TypedCallableIncompatibleSingature(self, n,
                                                                                                 caller,
                                                                                                 self.ctx),
                                                              args, kwargs)

        ret = self.fn(*args, **kwargs)
        if isinstance(self.inner, WrappedFunction):
            ret = self.inner.wrap_return(ret, bind2, TypedCallableReturnExecutionContext(self.ctx, self, True))

        ret = self.wrap_return(ret, bindings, TypedCallableReturnExecutionContext(self.ctx, self, False))
        return ret

    def get_original(self):
        return self.inner

    def wrap_arguments(self, ctxprv: WrappedFunctionContextProvider, args, kwargs):
        new_args = []
        i = 0
        for (arg, checker) in zip(args, self.argument_checker):
            res = checker.check_and_wrap(arg, ctxprv(i))
            new_args.append(res)
            i += 1

        if len(kwargs) > 0:
            raise self.ctx.wrap(UntypyTypeError(
                kwargs,
                self.describe()
            ).with_note("Keyword arguments are not supported in callable types."))

        bindings = None
        return new_args, kwargs, bindings

    def wrap_return(self, ret, bindings, ctx: ExecutionContext):
        return self.return_checker.check_and_wrap(ret, ctx)

    def describe(self) -> str:
        arguments = ", ".join(map(lambda e: e.describe(), self.argument_checker))
        return f"Callable[[{arguments}], {self.return_checker.describe()}]"

    def checker_for(self, name: str) -> TypeChecker:
        if name == 'return':
            return self.return_checker
        spec = inspect.getfullargspec(self.fn)
        try:
            ix = spec.args.index(name)
        except ValueError:
            raise self.ctx.wrap(UntypyAttributeError(
                f"No annotation found for argument {name}"
            ))
        if ix >= len(self.argument_checker):
            raise self.ctx.wrap(UntypyAttributeError(
                f"No annotation found for argument {name}"
            ))
        return self.argument_checker[ix]


class TypedCallableIncompatibleSingature(ExecutionContext):

    def __init__(self, tc: TypedCallable, arg_name: str, caller, upper: ExecutionContext):
        self.arg_name = arg_name
        self.tc = tc
        self.upper = upper
        self.caller = caller

    def wrap(self, err: UntypyTypeError) -> UntypyTypeError:
        (original_expected, _ind) = err.next_type_and_indicator()
        err = ArgumentExecutionContext(self.tc.inner, None, self.arg_name).wrap(err)

        func_decl = WrappedFunction.find_location(self.tc.inner)

        name = WrappedFunction.find_original(self.tc.inner).__name__

        (decl, ind) = err.next_type_and_indicator()
        err = err.with_frame(Frame(
            decl,
            ind,
            declared=None,
            responsable=func_decl
        ))

        previous_chain = UntypyTypeError(
            f"def {name}{self.tc.inner.describe()}",
            f"{self.tc.describe()}"
        )
        (decl, ind) = previous_chain.next_type_and_indicator()
        previous_chain = previous_chain.with_frame(Frame(
            decl,
            ind,
            declared=func_decl,
            responsable=None
        ))

        err = err.with_note(
            f"Argument '{self.arg_name}' of method '{name}' violates the signature of {self.tc.describe()}.")

        previous_chain = self.upper.wrap(previous_chain)

        return err.with_previous_chain(previous_chain)


class TypedCallableReturnExecutionContext(ExecutionContext):
    upper: ExecutionContext
    fn: TypedCallable

    def __init__(self, upper: ExecutionContext, fn: TypedCallable, invert: bool):
        self.upper = upper
        self.fn = fn
        self.invert = invert

    def wrap(self, err: UntypyTypeError) -> UntypyTypeError:
        desc = lambda s: s.describe()
        front_str = f"Callable[[{', '.join(map(desc, self.fn.argument_checker))}], "
        responsable = WrappedFunction.find_location(self.fn.inner)

        if self.invert:
            (next_ty, indicator) = err.next_type_and_indicator()
            err = ReturnExecutionContext(self.fn.inner).wrap(err)
            err = err.with_frame(Frame(
                f"{front_str}{next_ty}]",
                (" " * len(front_str)) + indicator,
                declared=None,
                responsable=responsable
            ))
            err = err.with_inverted_responsibility_type()
            return self.upper.wrap(err)

        (next_ty, indicator) = err.next_type_and_indicator()
        err = err.with_frame(Frame(
            f"{front_str}{next_ty}]",
            (" " * len(front_str)) + indicator,
            declared=None,
            responsable=responsable
        ))

        return self.upper.wrap(err)


class TypedCallableArgumentExecutionContext(ExecutionContext):
    upper: ExecutionContext
    fn: TypedCallable
    stack: inspect.FrameInfo
    argument_num: int

    def __init__(self, fn: TypedCallable, stack: inspect.FrameInfo, argument_num: int, upper: ExecutionContext):
        self.fn = fn
        self.stack = stack
        self.argument_num = argument_num
        self.upper = upper

    def declared_and_indicator(self, err: UntypyTypeError) -> Tuple[str, str]:
        (next_ty, indicator) = err.next_type_and_indicator()
        front_types = []
        back_types = []
        for n, ch in enumerate(self.fn.argument_checker):
            if n < self.argument_num:
                front_types.append(ch.describe())
            elif n > self.argument_num:
                back_types.append(ch.describe())

        l = len(f"Callable[[{', '.join(front_types)}")
        if len(front_types) > 0:
            l += len(', ')

        return f"Callable[[{', '.join(front_types + [next_ty] + back_types)}], {self.fn.return_checker.describe()}]", \
               (" " * l) + indicator

    def wrap(self, err: UntypyTypeError) -> UntypyTypeError:
        (type_declared, indicator_line) = self.declared_and_indicator(err)

        declared = WrappedFunction.find_location(self.fn.inner)

        responsable = Location.from_stack(self.stack)

        frame = Frame(
            type_declared,
            indicator_line,
            declared=declared,
            responsable=responsable,
        )

        err = err.with_frame(frame)
        err = err.with_inverted_responsibility_type()
        return self.upper.wrap(err)



from typing import TypeVar, Generic, Optional

I = TypeVar("I")


class Sequence(Generic[I]):
    # See https://docs.python.org/3/library/collections.abc.html

    def __getitem__(self, key: int) -> I:
        pass

    def __len__(self) -> int:
        pass

    def __contains__(self, key: I) -> bool:
        pass

    def index(self, value: I, start: Optional[int] = None, stop: Optional[int] = None) -> int:
        pass

    def count(self, value: I) -> int:
        pass

    # For these Methods: `iter` and `reversed` will fallback to `__getitem__` and `__len__`
    # This is just an optimization for some types.
    # Maybe for a future feature.

    # def __iter__(self) -> Iterator[I]:
    #    pass

    # def __reversed__(self) -> Iterator[I]:
    #    pass



from typing import Any, Optional, NoReturn

from untypy.error import UntypyTypeError
from untypy.interfaces import TypeChecker, TypeCheckerFactory, CreationContext, ExecutionContext


class NoneFactory(TypeCheckerFactory):

    def create_from(self, annotation: Any, ctx: CreationContext) -> Optional[TypeChecker]:
        if annotation is None or annotation is type(None) or annotation == NoReturn:
            return NoneChecker()
        else:
            return None


class NoneChecker(TypeChecker):
    def check_and_wrap(self, arg: Any, ctx: ExecutionContext) -> Any:
        if arg is None:
            return arg
        else:
            raise ctx.wrap(UntypyTypeError(arg, self.describe()))

    def describe(self) -> str:
        return "None"

    def base_type(self) -> list[Any]:
        return [None]



import typing
from collections.abc import Iterable as ABCIterable
from collections.abc import Sequence as ABCSequence
from typing import Optional, Any

from untypy.error import UntypyAttributeError, UntypyTypeError, Location, Frame, NO_GIVEN
from untypy.impl.interfaces.iterable import Iterable
from untypy.impl.interfaces.sequence import Sequence
from untypy.impl.interfaces.dict import Dict, DictLike
from untypy.impl.interfaces.list import List
from untypy.impl.interfaces.set import Set
from untypy.impl.protocol import ProtocolChecker
from untypy.impl.wrappedclass import WrappedType
from untypy.interfaces import TypeCheckerFactory, TypeChecker, CreationContext, ExecutionContext
from untypy.util import ReplaceTypeExecutionContext

InterfaceMapping = {
    dict: (Dict,),
    typing.Dict: (Dict,),
    DictLike: (DictLike,),
    list: (List,),
    typing.List: (List,),
    set: (Set,),
    typing.Set: (Set,),
    ABCIterable: (Iterable,),
    typing.Iterable: (Iterable,),
    ABCSequence: (Sequence,),
    typing.Sequence: (Sequence,)
}

class InterfaceFactory(TypeCheckerFactory):

    def create_from(self, annotation: Any, ctx: CreationContext, omit_tyargs=False) -> Optional[TypeChecker]:
        if annotation in InterfaceMapping:
            # Assume Any if no parameters are given
            (protocol,) = InterfaceMapping[annotation]
            bindings = protocol.__parameters__
            if len(bindings) == 0:
                raise AssertionError(f"This is a BUG. {annotation} has no generic params.")

            anys = (Any, ) * len(bindings)
            # handle Python inconsistency
            if hasattr(annotation, '__class_getitem__'):
                return self.create_from(
                    annotation.__class_getitem__(anys),
                    ctx,
                    omit_tyargs=True
                )
            elif hasattr(annotation, '__getitem__'):
                return self.create_from(
                    annotation.__getitem__(anys),
                    ctx,
                    omit_tyargs=True
                )

        elif hasattr(annotation, '__origin__') and hasattr(annotation,
                                                           '__args__') and annotation.__origin__ in InterfaceMapping:
            (protocol,) = InterfaceMapping[annotation.__origin__]
            bindings = protocol.__parameters__  # args of Generic super class
            origin = annotation.__origin__
            inner_checkers = []
            for param in annotation.__args__:
                ch = ctx.find_checker(param)
                if ch is None:
                    raise UntypyAttributeError(f"Could not resolve annotation {param} inside of {annotation}")
                inner_checkers.append(ch)
            if len(inner_checkers) != len(bindings):
                raise UntypyAttributeError(f"Expected {len(bindings)} type arguments inside of {annotation}")

            if omit_tyargs:
                name = f"{origin.__name__}"
            else:
                name = f"{origin.__name__}[" + (', '.join(map(lambda t: t.describe(), inner_checkers))) + "]"

            bindings = dict(zip(bindings, annotation.__args__))
            ctx = ctx.with_typevars(bindings)

            if type(origin) == type:
                template = WrappedType(protocol, ctx.with_typevars(bindings), name=name, implementation_template=origin,
                                       declared=ctx.declared_location(), overwrites=protocol)
                return InterfaceChecker(origin, template, name, ctx.declared_location())
            else:
                # type(origin) == collection.abc.ABCMeta
                return ProtocolChecker(protocol, ctx, altname=name, omit_tyargs=omit_tyargs)

        # Non Generic
        elif annotation in InterfaceMapping:
            (protocol,) = InterfaceMapping[annotation]

            # Edge-Case `TypingSequence has no __name__, like every other class`
            if annotation == typing.Sequence:
                name = 'Sequence'
            else:
                name = annotation.__name__

            if type(annotation) == type:
                template = WrappedType(protocol, ctx, name=name, implementation_template=annotation,
                                       declared=ctx.declared_location(), overwrites=protocol)
                return InterfaceChecker(annotation, template, name, ctx.declared_location())
            else:
                # type(origin) == collection.abc.ABCMeta
                return ProtocolChecker(protocol, ctx, altname=name, omit_tyargs=omit_tyargs)
        else:
            return None


class InterfaceChecker(TypeChecker):

    def __init__(self, origin, template, name, declared):
        self.origin = origin
        self.template = template
        self.name = name
        self.declared = declared
        pass

    def may_change_identity(self) -> bool:
        return True

    def check_and_wrap(self, arg: Any, ctx: ExecutionContext) -> Any:
        if not issubclass(type(arg), self.origin):
            raise ctx.wrap(UntypyTypeError(arg, self.describe()))

        if hasattr(arg, '_WrappedClassFunction__inner'):
            # Prevent Double Wrapping
            arg = arg._WrappedClassFunction__inner

        instance = self.template.__new__(self.template)
        instance._WrappedClassFunction__inner = arg
        instance._WrappedClassFunction__ctx = InterfaceCheckerContext(ctx, arg, self.name, self.declared)
        instance._WrappedClassFunction__return_ctx = ReplaceTypeExecutionContext(ctx, self.name)
        return instance

    def describe(self) -> str:
        return self.name


class InterfaceCheckerContext(ExecutionContext):
    def __init__(self, upper: ExecutionContext, arg: Any, name, declared: Location):
        self.upper = upper
        self.arg = arg
        self.name = name
        self.declared = declared

    def wrap(self, err: UntypyTypeError) -> UntypyTypeError:
        pv = self.upper.wrap(UntypyTypeError(
            given=NO_GIVEN,
        ).with_frame(Frame(
            type_declared=self.name,
            indicator_line="^" * len(self.name),
            responsable=None,
            declared=self.declared,
        )))

        return err.with_previous_chain(pv)



import ast
import inspect
import sys
from types import ModuleType
from typing import Optional, Any, Union, Callable

from .patching import wrap_function, patch_class, wrap_class, DefaultConfig
from .patching.ast_transformer import UntypyAstTransformer, did_no_code_run_before_untypy_enable, \
    UntypyAstImportTransformer
from .patching.import_hook import install_import_hook
from .patching.standalone_checker import StandaloneChecker
from .util.condition import FunctionCondition
from .util.return_traces import ReturnTracesTransformer, before_return, GlobalReturnTraceManager
from .util.tranformer_combinator import TransformerCombinator

GlobalConfig = DefaultConfig

"""
This function is called before any return statement, to store which was the last return.
For this the AST is transformed using ReturnTracesTransformer.
Must be in untypy so it can be used in transformed module.
Must also be in other module, so it can be used from inside (No circular imports).
"""
_before_return = before_return

_importhook_transformer_builder = lambda path, file: TransformerCombinator(UntypyAstTransformer(),
                                                                           ReturnTracesTransformer(file))

def just_install_hook(prefixes=[]):
    global GlobalConfig

    def predicate(module_name):
        for p in prefixes:
            if module_name == p:
                return True
            elif module_name.startswith(p + "."):
                return True
        return False

    GlobalConfig = DefaultConfig._replace(checkedprefixes=[*prefixes])
    install_import_hook(predicate, _importhook_transformer_builder)


def transform_tree(tree, file):
    UntypyAstTransformer().visit(tree)
    ReturnTracesTransformer(file).visit(tree)
    ast.fix_missing_locations(tree)


def enable(*, recursive: bool = True, root: Union[ModuleType, str, None] = None, prefixes: list[str] = []) -> None:
    global GlobalConfig
    caller = _find_calling_module()
    exit_after = False

    if root is None:
        root = caller
        exit_after = True
    if caller is None:
        raise Exception("Couldn't find loading Module. This is a Bug.")

    rootname = root
    if hasattr(rootname, '__name__'):
        rootname = root.__name__

    GlobalConfig = DefaultConfig._replace(checkedprefixes=[rootname])

    def predicate(module_name):
        if recursive:
            # Patch if Submodule
            if module_name.startswith(rootname + "."):
                return True
            else:
                for p in prefixes:
                    if module_name == p:
                        return True
                    elif module_name.startswith(p + "."):
                        return True
            return False
        else:
            raise AssertionError("You cannot run 'untypy.enable()' twice!")

    transformer = _importhook_transformer_builder
    install_import_hook(predicate, transformer)
    _exec_module_patched(root, exit_after, transformer(caller.__name__.split("."), caller.__file__))


def enable_on_imports(*prefixes):
    print("!!!! THIS FEATURE HAS UNEXPECTED SIDE EFFECTS WHEN IMPORTING ABSOLUTE SUBMODULES !!!")
    # TODO: Fix import of submodules should not change parent module.
    global GlobalConfig
    GlobalConfig = DefaultConfig._replace(checkedprefixes=[*prefixes])
    caller = _find_calling_module()

    def predicate(module_name: str):
        module_name = module_name.replace('__main__.', '')
        for p in prefixes:
            if module_name == p:
                return True
            elif module_name.startswith(p + "."):
                return True
            else:
                return False

    transformer = _importhook_transformer_builder
    install_import_hook(predicate, transformer)
    _exec_module_patched(caller, True, transformer(caller.__name__.split("."), caller.__file__))


def _exec_module_patched(mod: ModuleType, exit_after: bool, transformer: ast.NodeTransformer):
    source = inspect.getsource(mod)
    tree = compile(source, mod.__file__, 'exec', ast.PyCF_ONLY_AST,
                   dont_inherit=True, optimize=-1)

    if not did_no_code_run_before_untypy_enable(tree):
        raise AssertionError("Please put 'untypy.enable()' at the start of your module like so:\n"
                             "\timport untypy\n"
                             "\tuntypy.enable()")

    transformer.visit(tree)
    ReturnTracesTransformer(mod.__file__).visit(tree)
    ast.fix_missing_locations(tree)
    patched_mod = compile(tree, mod.__file__, 'exec', dont_inherit=True, optimize=-1)
    stack = list(map(lambda s: s.frame, inspect.stack()))
    try:
        exec(patched_mod, mod.__dict__)
    except Exception as e:
        while e.__traceback__.tb_frame in stack:
            e.__traceback__ = e.__traceback__.tb_next
        sys.excepthook(type(e), e, e.__traceback__)
        if exit_after: sys.exit(-1)
    if exit_after: sys.exit(0)


def _find_calling_module() -> Optional[ModuleType]:
    for caller in inspect.stack():
        if caller.filename != __file__:
            mod = inspect.getmodule(caller.frame)
            if mod is not None:
                return mod
    return None


def _condgetfc(func):
    if hasattr(func, "__fc"):
        return getattr(func, "__fc")
    else:
        fc = FunctionCondition()
        setattr(func, "__fc", fc)
        fc.func = func
        return fc


def precondition(cond):
    def decorator(func):
        fc = _condgetfc(func)
        fc.precondition.append(cond)
        return func

    return decorator


def postcondition(cond):
    def decorator(func):
        fc = _condgetfc(func)
        fc.postcondition.append(cond)
        return func

    return decorator


def unchecked(fn):
    setattr(fn, "__unchecked", True)
    return fn


def patch(a: Any) -> Any:
    global GlobalConfig
    if hasattr(a, '__unchecked'):
        return a

    if inspect.isfunction(a):
        return wrap_function(a, GlobalConfig)
    elif inspect.isclass(a):
        patch_class(a, GlobalConfig)
        return a
    else:
        return a


typechecked = patch


def wrap_import(a: Any) -> Any:
    global GlobalConfig
    if inspect.isfunction(a):
        return wrap_function(a, GlobalConfig)
    elif inspect.isclass(a) or inspect.ismodule(a):
        return wrap_class(a, GlobalConfig)
    else:
        return a


def checker(annotation: Callable[[], Any], location: Any, ctx=None) -> Callable[[Any], Any]:
    """
    :param annotation: A functions that returns the annotation lazily. (Needed for FowardRefs)
    :param location: Where was it declared?
    :return: A type checker function
    """
    return StandaloneChecker(annotation, location, DefaultConfig, ctx)



from typing import Optional, Any, Iterator

from untypy.error import UntypyTypeError, UntypyAttributeError, Frame
from untypy.interfaces import TypeCheckerFactory, TypeChecker, ExecutionContext, CreationContext


class AnnotatedChecker:
    def check(self, arg: Any, ctx: ExecutionContext) -> None:
        pass


class AnnotatedCheckerCallable(AnnotatedChecker):
    def __init__(self, annotated, callable):
        self.callable = callable
        self.annotated = annotated

    def check(self, arg: Any, ctx: ExecutionContext) -> None:
        res = self.callable(arg)
        if not res:
            exp = self.annotated.describe()
            # raise error on falsy value
            err = UntypyTypeError(
                given=arg,
                expected=exp
            )
            if self.annotated.is_anonymous:
                err = err.with_note(f"condition in {exp} does not hold")
            (t, i) = err.next_type_and_indicator()
            for info in self.annotated.info:
                err = err.with_note("    - " + info)
            raise ctx.wrap(err)


class AnnotatedCheckerContainer(AnnotatedChecker):
    def __init__(self, annotated, cont):
        self.cont = cont
        self.annotated = annotated

    def check(self, arg: Any, ctx: ExecutionContext) -> None:
        if arg not in self.cont:
            # raise error on falsy value
            err = UntypyTypeError(
                given=arg,
                expected=self.annotated.describe()
            )
            if self.annotated.is_anonymous:
                err = err.with_note(f"{repr(arg)} is not in {repr(self.cont)}.")

            for info in self.annotated.info:
                err = err.with_note("    - " + info)

            raise ctx.wrap(err)


class AnnotatedFactory(TypeCheckerFactory):

    def create_from(self, annotation: Any, ctx: CreationContext) -> Optional[TypeChecker]:
        if hasattr(annotation, '__metadata__') and hasattr(annotation, '__origin__'):
            inner = ctx.find_checker(annotation.__origin__)
            if inner is None:
                raise ctx.wrap(UntypyAttributeError(f"Could not resolve annotation "
                                                    f"'{repr(annotation.__origin__)}' "
                                                    f"inside of '{annotation}'."))

            return AnnotatedChecker(annotation, inner, annotation.__metadata__, ctx)
        else:
            return None


class AnnotatedChecker(TypeChecker):
    def __init__(self, annotated, inner: TypeChecker, metadata: Iterator, ctx: CreationContext):
        self.annotated = annotated
        self.inner = inner

        meta = []
        info = []
        for m in metadata:
            if callable(m):
                meta.append(AnnotatedCheckerCallable(self, m))
            elif isinstance(m, str):
                info.append(m)
            elif hasattr(m, '__contains__'):
                meta.append(AnnotatedCheckerContainer(self, m))
            else:
                raise ctx.wrap(UntypyAttributeError(f"Unsupported metadata '{repr(m)}' in '{repr(self.annotated)}'.\n"
                                                    f"Only callables or objects providing __contains__ are allowed."))
        self.meta = meta
        self.info = info
        if len(info) == 1:
            self.name = info[0]
            self.info = []
        else:
            self.name = None

    def check_and_wrap(self, arg: Any, ctx: ExecutionContext) -> Any:
        wrapped = self.inner.check_and_wrap(arg, AnnotatedCheckerExecutionContext(self, ctx))
        for ck in self.meta:
            ck.check(wrapped, ctx)
        return wrapped

    def describe(self) -> str:
        if self.name:
            return self.name
        elif len(self.info) > 0:
            text = ", ".join(map(lambda a: f"'{a}'", self.info))
            return f"Annotated[{text}]"
        else:
            return repr(self.annotated)

    def base_type(self):
        return self.inner.base_type()

    @property
    def is_anonymous(self):
        return self.name is None


class AnnotatedCheckerExecutionContext(ExecutionContext):
    def __init__(self, ch: AnnotatedChecker, upper: ExecutionContext):
        self.ch = ch
        self.upper = upper

    def wrap(self, err: UntypyTypeError) -> UntypyTypeError:
        offset = self.ch.describe().find("[") + 1

        (t, i) = err.next_type_and_indicator()
        if self.ch.name is not None:
            i = "^" * len(self.ch.describe())

        err = err.with_frame(Frame(
            type_declared=self.ch.describe(),
            indicator_line=(" " * offset) + i,
            declared=None,
            responsable=None
        ))

        return self.upper.wrap(err)



import inspect
import sys
from types import ModuleType
from typing import Any, Callable, Union, Optional

from untypy.error import Location, UntypyAttributeError, UntypyTypeError
from untypy.impl.any import SelfChecker, AnyChecker
from untypy.interfaces import TypeChecker, CreationContext, ExecutionContext, WrappedFunction, \
    WrappedFunctionContextProvider
from untypy.util import ArgumentExecutionContext, ReturnExecutionContext


def find_signature(member, ctx: CreationContext):
    signature = inspect.signature(member)

    checkers = {}
    for key in signature.parameters:
        if key == 'self':
            checkers[key] = SelfChecker()
        else:
            param = signature.parameters[key]
            if param.annotation is inspect.Parameter.empty:
                checkers[key] = AnyChecker()
                continue

            checker = ctx.find_checker(param.annotation)
            if checker is None:
                checkers[key] = AnyChecker()
                continue
            checkers[key] = checker

    if signature.return_annotation is inspect.Parameter.empty:
        checkers['return'] = AnyChecker()
    else:
        return_checker = ctx.find_checker(signature.return_annotation)
        if return_checker is None:
            raise ctx.wrap(UntypyAttributeError(f"\n\tUnsupported Type Annotation: {signature.return_annotation}\n"
                                                f"for Return Value of function {member.__name__}\n"))
        checkers['return'] = return_checker
    return signature, checkers


def wrap_arguments(signature: inspect.signature, checker: dict[str, TypeChecker],
                   ctxprv: WrappedFunctionContextProvider, args, kwargs):
    bindings = signature.bind(*args, **kwargs)
    bindings.apply_defaults()
    for name in bindings.arguments:
        check = checker[name]
        ctx = ctxprv(name)
        bindings.arguments[name] = check.check_and_wrap(bindings.arguments[name], ctx)
    return bindings.args, bindings.kwargs


def WrappedType(template: Union[type, ModuleType], ctx: CreationContext, *,
                implementation_template: Union[type, ModuleType, None] = None,
                create_type: Optional[type] = None,
                name: Optional[str] = None,
                declared: Optional[Location] = None,
                overwrites: Optional[type] = None):
    blacklist = ['__class__', '__delattr__', '__dict__', '__dir__',
                 '__doc__', '__getattribute__', '__getattr__', '__init_subclass__',
                 '__new__', '__setattr__', '__subclasshook__', '__weakref__']

    create_fn = None
    if create_type is not None:
        create_fn = lambda: create_type.__new__(create_type)

    if type(template) is type and create_type is None:
        create_fn = lambda: template.__new__(template)

    if implementation_template is None:
        implementation_template = template

    if create_fn is None:
        def raise_err():
            raise TypeError("This is not a Callable")

        create_fn = raise_err

    list_of_attr = dict()

    for attr in dir(template):
        if attr in blacklist:
            continue

        if overwrites is not None and hasattr(overwrites, attr):
            a = getattr(overwrites, attr)
            if hasattr(a, '__overwrite') and getattr(a, '__overwrite') == 'simple':
                list_of_attr[attr] = getattr(overwrites, attr)
                continue

        original = getattr(template, attr)
        if type(original) == type:  # Note: Order matters, types are also callable
            if type(template) is type:
                list_of_attr[attr] = WrappedType(original, ctx)

        elif callable(original):
            try:
                (signature, checker) = find_signature(original, ctx)
                implementation_fn = getattr(implementation_template, attr)
                if implementation_fn is not None:
                    if overwrites is not None and hasattr(overwrites, attr) and hasattr(getattr(overwrites, attr),
                                                                                        '__overwrite') and getattr(
                            getattr(overwrites, attr), '__overwrite') == 'advanced':
                        list_of_attr[attr] = WrappedClassFunction(implementation_fn, signature, checker,
                                                                  create_fn=create_fn,
                                                                  declared=declared).build_overwrite(
                            getattr(overwrites, attr), ctx)
                    else:
                        list_of_attr[attr] = WrappedClassFunction(implementation_fn, signature, checker,
                                                                  create_fn=create_fn, declared=declared).build()
            except ValueError as e:
                # this fails sometimes on built-ins.
                # "ValueError: no signature found for builtin"
                pass
    out = None
    if type(template) is type:
        if name is None:
            name = f"{template.__name__}Wrapped"
        out = type(name, (template,), list_of_attr)
    elif inspect.ismodule(template):
        out = type("WrappedModule", (), list_of_attr)

    return out


class WrappedClassFunction(WrappedFunction):
    def __init__(self,
                 inner: Callable,
                 signature: inspect.Signature,
                 checker: dict[str, TypeChecker], *,
                 create_fn: Callable[[], Any],
                 declared: Optional[Location] = None):
        self.inner = inner
        self.signature = signature
        self.checker = checker
        self.create_fn = create_fn
        self._declared = declared
        self.fc = None
        if hasattr(self.inner, "__fc"):
            self.fc = getattr(self.inner, "__fc")

    def build_overwrite(self, f, ctx: CreationContext):
        fn = self.inner

        w = f(self, ctx)

        setattr(w, '__wrapped__', fn)
        setattr(w, '__name__', fn.__name__)
        setattr(w, '__signature__', self.signature)
        setattr(w, '__wf', self)
        return w

    def build(self):
        fn = self.inner
        name = fn.__name__

        def wrapper_cls(*args, **kwargs):
            caller = sys._getframe(1)
            (args, kwargs, bindings) = self.wrap_arguments(
                lambda n: ArgumentExecutionContext(wrapper_cls, caller, n, declared=self.declared()),
                args, kwargs)
            ret = fn(*args, **kwargs)
            return self.wrap_return(ret, bindings, ReturnExecutionContext(self))

        def wrapper_self(me, *args, **kwargs):
            if name == '__init__':
                me.__return_ctx = None
                me.__ctx = None
                me.__inner = self.create_fn()
            caller = sys._getframe(1)
            (args, kwargs, bindings) = self.wrap_arguments(
                lambda n: ArgumentExecutionContext(wrapper_self, caller, n, declared=self.declared(), upper=me.__ctx),
                (me.__inner, *args), kwargs)
            ret = fn(*args, **kwargs)
            if me.__return_ctx is None:
                return self.wrap_return(ret, bindings, ReturnExecutionContext(self))
            else:
                return self.wrap_return(ret, bindings, me.__return_ctx)

        if inspect.iscoroutine(self.inner):
            raise UntypyAttributeError("Async Functions are currently not supported.")
        else:
            if 'self' in self.checker:
                w = wrapper_self
            else:
                w = wrapper_cls

        setattr(w, '__wrapped__', fn)
        setattr(w, '__name__', fn.__name__)
        setattr(w, '__signature__', self.signature)
        setattr(w, '__wf', self)
        return w

    def get_original(self):
        return self.inner

    def wrap_arguments(self, ctxprv: WrappedFunctionContextProvider, args, kwargs):
        try:
            bindings = self.signature.bind(*args, **kwargs)
        except TypeError as e:
            err = UntypyTypeError(header=str(e))
            raise ctxprv("").wrap(err)

        bindings.apply_defaults()
        if self.fc is not None:
            self.fc.prehook(bindings, ctxprv)
        for name in bindings.arguments:
            check = self.checker[name]
            ctx = ctxprv(name)
            bindings.arguments[name] = check.check_and_wrap(bindings.arguments[name], ctx)
        return bindings.args, bindings.kwargs, bindings

    def wrap_return(self, ret, bindings, ctx: ExecutionContext):
        check = self.checker['return']
        if self.fc is not None:
            self.fc.posthook(ret, bindings, ctx)
        return check.check_and_wrap(ret, ctx)

    def describe(self) -> str:
        fn = WrappedFunction.find_original(self.inner)
        return f"{fn.__name__}" + str(self.signature)

    def checker_for(self, name: str) -> TypeChecker:
        return self.checker[name]

    def declared(self) -> Location:
        if self._declared is None:
            return WrappedFunction.find_location(self.inner)
        else:
            return self._declared



import ast


class TransformerCombinator(ast.NodeTransformer):

    def __init__(self, *inner: ast.NodeTransformer):
        self.inner = inner

    def visit(self, node):
        for inner in self.inner:
            inner.visit(node)



import inspect
from typing import Any, Optional, TypeVar, List, Dict
import typing

from untypy.interfaces import CreationContext, TypeChecker
from .annotated import AnnotatedFactory
from .any import AnyFactory
from .callable import CallableFactory
from .dummy_delayed import DummyDelayedFactory
from .generator import GeneratorFactory
from .generic import GenericFactory
from .interface import InterfaceFactory
from .iterator import IteratorFactory
from .literal import LiteralFactory
from .none import NoneFactory
from .optional import OptionalFactory
from .protocol import ProtocolFactory
from .simple import SimpleFactory
from .string_forward_refs import StringForwardRefFactory
from .tuple import TupleFactory
from .union import UnionFactory
from ..error import Location, UntypyAttributeError

# More Specific Ones First
_FactoryList = [
    AnyFactory(),
    NoneFactory(),
    AnnotatedFactory(),
    ProtocolFactory(),  # must be higher then Generic
    GenericFactory(),
    CallableFactory(),
    LiteralFactory(),
    OptionalFactory(),  # must be higher then Union
    UnionFactory(),
    TupleFactory(),
    DummyDelayedFactory(),
    GeneratorFactory(),
    IteratorFactory(),
    InterfaceFactory(),
    StringForwardRefFactory(),  # resolve types passed as strings
    # must come last
    SimpleFactory()
]


class DefaultCreationContext(CreationContext):

    def __init__(self, typevars: Dict[TypeVar, Any], declared_location: Location,
                 checkedpkgprefixes: List[str], eval_context: Optional[Any] = None):
        self.typevars = typevars
        self.declared = declared_location
        self.checkedpkgprefixes = checkedpkgprefixes
        self._eval_context = eval_context

    def declared_location(self) -> Location:
        return self.declared

    def find_checker(self, annotation: Any) -> Optional[TypeChecker]:
        for fac in _FactoryList:
            res = fac.create_from(annotation=annotation, ctx=self)
            if res is not None:
                return res
        return None

    def wrap(self, err: UntypyAttributeError) -> UntypyAttributeError:
        return err.with_location(self.declared)

    def resolve_typevar(self, var: TypeVar) -> typing.Tuple[bool, Any]:
        # Not result may be None
        if var in self.typevars:
            return True, self.typevars[var]
        else:
            return False, None

    def all_typevars(self) -> List[TypeVar]:
        return list(self.typevars.keys())

    def with_typevars(self, typevars: Dict[TypeVar, Any]) -> CreationContext:
        tv = self.typevars.copy()
        tv.update(typevars)
        return DefaultCreationContext(tv, self.declared, self.checkedpkgprefixes, self._eval_context)

    def should_be_inheritance_checked(self, annotation: type) -> bool:
        m = inspect.getmodule(annotation)

        for pkgs in self.checkedpkgprefixes:
            # Inheritance should be checked on types
            # when the type's module or its parent lies in the "user code".
            # Inheritance of types of extern modules should be not be checked.
            if m.__name__ == pkgs or m.__name__.startswith(pkgs + "."):
                return True

        return False

    def eval_context(self):
        return self._eval_context



from typing import Iterator, Protocol, TypeVar, Generic, Optional, Iterable, Tuple, Any

A = TypeVar("A")
B = TypeVar("B")


class DictLike(Protocol[A, B]):
    """
    This protocol implements a subset of dict.
    It exists solly to prevent an recursion issue
    inside of WDict
    """

    def __iter__(self) -> Iterator[A]:
        pass

    def __getitem__(self, key: A) -> B:
        pass


K = TypeVar("K")
V = TypeVar("V")


# See: https://docs.python.org/3/library/stdtypes.html#typesmapping
class Dict(Generic[K, V], dict):
    def clear(self) -> None:
        pass

    # Cannot Typecheck Copy -> Leads to endless recursion in "UntypyInterfaces"
    # def copy(self) -> dict[K,V]:
    #     pass

    def get(self, key: K, default: Optional[V] = None) -> Optional[V]:
        pass

    def items(self) -> Iterable[Tuple[K, V]]:
        pass

    def keys(self) -> Iterable[K]:
        pass

    def pop(self, k: K, default: Optional[V] = None) -> Optional[V]:
        pass

    def popitem(self) -> Tuple[K, V]:
        pass

    # Miss-match See: https://github.com/skogsbaer/write-your-python-program/issues/19
    def setdefault(self, key: K, default: V) -> V:
        pass

    def update(self, *E: Iterable[DictLike[K, V]], **F: Optional[DictLike[K, V]]) -> Any:
        pass

    def values(self) -> Iterable[V]:
        pass

    def __contains__(self, key: K) -> bool:
        pass

    def __delitem__(self, key: K) -> None:
        pass

    def __iter__(self) -> Iterator[K]:
        pass

    def __len__(self) -> int:
        pass

    # Untypy does not support generic functions :/
    # def __or__(self, other : dict[I, J]) -> dict[Union[K,I], Union[V,J]]:
    #     pass

    def __reversed__(self) -> Iterator[K]:
        pass

    # Untypy does not support generic functions :/
    # def __ror__(self, other : dict[I, J]) -> dict[Union[K,I], Union[V,J]]:
    #     """ Return value|self. """
    #     pass

    def __getitem__(self, key: K) -> V:
        pass

    def __setitem__(self, key: K, value: V) -> None:
        pass



from typing import Any, Optional, Tuple

from untypy.error import UntypyTypeError, Frame
from untypy.interfaces import TypeChecker, TypeCheckerFactory, CreationContext, ExecutionContext
from untypy.util import CompoundTypeExecutionContext

TupleType = type(Tuple[str, int])
TupleTypeB = type(tuple[str, int])


class TupleFactory(TypeCheckerFactory):

    def create_from(self, annotation: Any, ctx: CreationContext) -> Optional[TypeChecker]:
        if (type(annotation) is TupleType or type(annotation) is TupleTypeB) and annotation.__origin__ == tuple:
            args = annotation.__args__
            if len(args) == 2 and args[1] == Ellipsis:
                checker = ctx.find_checker(args[0])
                if checker is None:
                    return None
                else:
                    return VariadicTupleChecker(checker)
            inner = []
            for arg in args:
                checker = ctx.find_checker(arg)
                if checker is None:
                    return None
                else:
                    inner.append(checker)

            return TupleChecker(inner)
        else:
            return None

class VariadicTupleChecker(TypeChecker):
    inner: TypeChecker

    def __init__(self, inner: TypeChecker):
        self.inner = inner

    def check_and_wrap(self, arg: Any, ctx: ExecutionContext) -> Any:
        if not type(arg) is tuple:
            raise ctx.wrap(UntypyTypeError(arg, self.describe()))
        out = []
        for elm in arg:
            out.append(self.inner.check_and_wrap(elm, VariadicTupleExecutionContext(ctx)))
        return tuple(out)

    def base_type(self) -> Any:
        return [tuple]

    def describe(self) -> str:
        desc = self.inner.describe()
        return f"tuple[{desc}, ...]"


class VariadicTupleExecutionContext(ExecutionContext):
    upper: ExecutionContext

    def __init__(self, upper: ExecutionContext):
        self.upper = upper

    def wrap(self, err: UntypyTypeError) -> UntypyTypeError:
        next_type, indicator = err.next_type_and_indicator()
        err = err.with_frame(Frame(
            f"tuple[{next_type}, ...]",
            (" " * len("tuple[") + indicator),
            None,
            None
        ))
        return self.upper.wrap(err)

class TupleChecker(TypeChecker):
    inner: list[TypeChecker]

    def __init__(self, inner: list[TypeChecker]):
        self.inner = inner

    def check_and_wrap(self, arg: Any, ctx: ExecutionContext) -> Any:
        if not type(arg) is tuple or len(arg) != len(self.inner):
            raise ctx.wrap(UntypyTypeError(arg, self.describe()))

        out = []
        idx = 0
        for elm, checker in zip(arg, self.inner):
            out.append(checker.check_and_wrap(elm, TupleExecutionContext(ctx, self.inner, idx)))
            idx += 1

        return tuple(out)

    def base_type(self) -> Any:
        out = []
        for checker in self.inner:
            out.append(checker.base_type())
        return tuple(out)

    def describe(self) -> str:
        desc = lambda s: s.describe()
        return f"tuple[{', '.join(map(desc, self.inner))}]"


class TupleExecutionContext(CompoundTypeExecutionContext):
    def name(self):
        return "tuple"



from __future__ import annotations as anns
import sys
import os
import os.path
import argparse
import json
import traceback
import shutil
import site
import importlib
import re
import code
import ast
from modulefinder import ModuleFinder
from pathlib import Path
import subprocess

__wypp_runYourProgram = 1

def die(ecode=1):
    if sys.flags.interactive:
        os._exit(ecode)
    else:
        sys.exit(ecode)

def getEnv(name, conv, default):
    s = os.getenv(name)
    if s is None:
        return default
    try:
        return conv(s)
    except:
        return default

VERBOSE = False # set via commandline
DEBUG = getEnv("WYPP_DEBUG", bool, False)

def enableVerbose():
    global VERBOSE
    VERBOSE = True

LIB_DIR = os.path.dirname(__file__)
INSTALLED_MODULE_NAME = 'wypp'
FILES_TO_INSTALL = ['writeYourProgram.py', 'drawingLib.py', '__init__.py']

UNTYPY_DIR = os.path.join(LIB_DIR, "..", "deps", "untypy", "untypy")
UNTYPY_MODULE_NAME = 'untypy'

def verbose(s):
    if VERBOSE or DEBUG:
        printStderr('[V] ' + str(s))

def printStderr(s=''):
    sys.stderr.write(s + '\n')

class InstallMode:
    dontInstall = 'dontInstall'
    installOnly = 'installOnly'
    install = 'install'
    assertInstall = 'assertInstall'
    allModes = [dontInstall, installOnly, install, assertInstall]

def parseCmdlineArgs(argList):
    parser = argparse.ArgumentParser(description='Run Your Program!',
                        formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('--check-runnable', dest='checkRunnable', action='store_const',
                        const=True, default=False,
                        help='Abort with exit code 1 if loading the file raises errors')
    parser.add_argument('--check', dest='check', action='store_const',
                        const=True, default=False,
                        help='Abort with exit code 1 if there are test errors.')
    parser.add_argument('--install-mode', dest='installMode', type=str,
                        default=InstallMode.dontInstall,
                        help="""The install mode can be one of the following:
- dontInstall     do not install the wypp library (default)
- installOnly     install the wypp library and quit
- install         install the wypp library and continue even if installation fails
- assertInstall   check whether wypp is installed
""")
    parser.add_argument('--verbose', dest='verbose', action='store_const',
                        const=True, default=False,
                        help='Be verbose')
    parser.add_argument('--debug', dest='debug', action='store_const',
                        const=True, default=False,
                        help='Enable debugging')
    parser.add_argument('--quiet', dest='quiet', action='store_const',
                        const=True, default=False, help='Be extra quiet')
    parser.add_argument('--no-clear', dest='noClear', action='store_const',
                        const=True, default=False, help='Do not clear the terminal')
    parser.add_argument('--test-file', dest='testFile',
                        type=str, help='Run additional tests contained in this file.')
    parser.add_argument('--change-directory', dest='changeDir', action='store_const',
                        const=True, default=False,
                        help='Change to the directory of FILE before running')
    parser.add_argument('--interactive', dest='interactive', action='store_const',
                        const=True, default=False,
                        help='Run REPL after the programm has finished')
    parser.add_argument('--no-typechecking', dest='checkTypes', action='store_const',
                        const=False, default=True,
                        help='Do not check type annotations')
    parser.add_argument('file', metavar='FILE',
                        help='The file to run', nargs='?')
    if argList is None:
        argList = sys.argv[1:]
    try:
        args, restArgs = parser.parse_known_args(argList)
    except SystemExit as ex:
        die(ex.code)
    if args.file and not args.file.endswith('.py'):
        printStderr(f'ERROR: file {args.file} is not a python file')
        die()
    if args.installMode not in InstallMode.allModes:
        printStderr(f'ERROR: invalid install mode {args.installMode}.')
        die()
    return (args, restArgs)

def readFile(path):
    try:
        with open(path, encoding='utf-8') as f:
            return f.read()
    except UnicodeDecodeError:
        with open(path) as f:
            return f.read()

def readGitVersion():
    thisDir = os.path.basename(LIB_DIR)
    baseDir = os.path.join(LIB_DIR, '..', '..')
    if thisDir == 'src' and os.path.isdir(os.path.join(baseDir, '.git')):
        try:
            h = subprocess.check_output(['git', '-C', baseDir, 'rev-parse', '--short', 'HEAD'],
                encoding='UTF-8').strip()
            changes = subprocess.check_output(
                    ['git', '-C', baseDir, 'status', '--porcelain', '--untracked-files=no'],
                    encoding='UTF-8').strip()
            if changes:
                return f'git-{h}-dirty'
            else:
                return f'git-{h}'
        except subprocess.CalledProcessError:
            return 'git-?'
    else:
        return None

def readVersion():
    version = readGitVersion()
    if version is not None:
        return version
    try:
        content = readFile(os.path.join(LIB_DIR, '..', '..', 'package.json'))
        d = json.loads(content)
        version = d['version']
    except:
        pass
    return version

def printWelcomeString(file, version, useUntypy):
    cwd = os.getcwd() + "/"
    if file.startswith(cwd):
        file = file[len(cwd):]
    versionStr = '' if not version else 'Version %s, ' % version
    pythonVersion = sys.version.split()[0]
    tycheck = ''
    if not useUntypy:
        tycheck = ', no typechecking'
    printStderr('=== WELCOME to "Write Your Python Program" ' +
                '(%sPython %s, %s%s) ===' % (versionStr, pythonVersion, file, tycheck))

def isSameFile(f1, f2):
    x = readFile(f1)
    y = readFile(f2)
    return x == y

def installFromDir(srcDir, targetDir, mod, files=None):
    verbose(f'Installing from {srcDir} to {targetDir}/{mod}')
    if files is None:
        files = [p.relative_to(srcDir) for p in Path(srcDir).rglob('*.py')]
    else:
        files = [Path(f) for f in files]
    installDir = os.path.join(targetDir, mod)
    os.makedirs(installDir, exist_ok=True)
    installedFiles = sorted([p.relative_to(installDir) for p in Path(installDir).rglob('*.py')])
    wantedFiles = sorted(files)
    if installedFiles == wantedFiles:
        for i in range(len(installedFiles)):
            f1 = os.path.join(installDir, installedFiles[i])
            f2 = os.path.join(srcDir, wantedFiles[i])
            if not isSameFile(f1, f2):
                verbose(f'{f1} and {f2} differ')
                break
        else:
            # no break, all files equal
            verbose(f'All files from {srcDir} already installed in {targetDir}/{mod}')
            return True
    else:
        verbose(f'Installed files {installedFiles} and wanted files {wantedFiles} are different')
    for f in installedFiles:
        p = os.path.join(installDir, f)
        os.remove(p)
    d = os.path.join(installDir, mod)
    for f in wantedFiles:
        src = os.path.join(srcDir, f)
        tgt = os.path.join(installDir, f)
        os.makedirs(os.path.dirname(tgt), exist_ok=True)
        shutil.copyfile(src, tgt)
    verbose(f'Finished installation from {srcDir} to {targetDir}/{mod}')
    return False

def installLib(mode):
    verbose("installMode=" + mode)
    if mode == InstallMode.dontInstall:
        verbose("No installation of WYPP should be performed")
        return
    targetDir = os.getenv('WYPP_INSTALL_DIR', site.USER_SITE)
    try:
        allEq1 = installFromDir(LIB_DIR, targetDir, INSTALLED_MODULE_NAME, FILES_TO_INSTALL)
        allEq2 = installFromDir(UNTYPY_DIR, targetDir, UNTYPY_MODULE_NAME)
        if allEq1 and allEq2:
            verbose(f'WYPP library in {targetDir} already up to date')
            if mode == InstallMode.installOnly:
                printStderr(f'WYPP library in {targetDir} already up to date')
            return
        else:
            if mode == InstallMode.assertInstall:
                printStderr("The WYPP library was not installed before running this command.")
                die(1)
            printStderr(f'The WYPP library has been successfully installed in {targetDir}')
    except Exception as e:
        printStderr('Installation of the WYPP library failed: ' + str(e))
        if mode == InstallMode.installOnly:
            raise e
    if mode == InstallMode.installOnly:
        printStderr('Exiting after installation of the WYPP library')
        die(0)

class Lib:
    def __init__(self, mod, properlyImported):
        self.properlyImported = properlyImported
        if not properlyImported:
            self.initModule = mod['initModule']
            self.resetTestCount = mod['resetTestCount']
            self.printTestResults = mod['printTestResults']
            self.dict = mod
        else:
            self.initModule = mod.initModule
            self.resetTestCount = mod.resetTestCount
            self.printTestResults = mod.printTestResults
            d = {}
            self.dict = d
            for name in dir(mod):
                if name and name[0] != '_':
                  d[name] = getattr(mod, name)

def prepareLib(onlyCheckRunnable):
    libDefs = None
    mod = INSTALLED_MODULE_NAME
    verbose('Attempting to import ' + mod)
    wypp = importlib.import_module(mod)
    libDefs = Lib(wypp, True)
    verbose('Successfully imported module ' + mod + ' from file ' + wypp.__file__)
    libDefs.initModule(enableChecks=not onlyCheckRunnable,
                       quiet=onlyCheckRunnable)
    return libDefs

importRe = importRe = re.compile(r'^import\s+wypp\s*$|^from\s+wypp\s+import\s+\*\s*$')

def findWyppImport(fileName):
    lines = []
    try:
        with open(fileName) as f:
            lines = f.readlines()
    except Exception as e:
        verbose('Failed to read code from %s: %s' % (fileName, e))
    for l in lines:
        if importRe.match(l):
            return True
    return False

def findImportedModules(path, file):
    finder = ModuleFinder(path=path)
    try:
        finder.run_script(file)
    except:
        return []
    realdirs = [os.path.realpath(p) for p in path]
    res = []
    for name, mod in finder.modules.items():
        if name != '__main__' and mod.__file__:
            realp = os.path.realpath(mod.__file__)
            good = False
            for d in realdirs:
                if realp.startswith(d):
                    good = True
                    break
            if good:
                res.append(name)
    return res

class RunSetup:
    def __init__(self, sysPath):
        self.sysPath = sysPath
        self.sysPathInserted = False
    def __enter__(self):
        if self.sysPath not in sys.path:
            sys.path.insert(0, self.sysPath)
            self.sysPathInserted = True
    def __exit__(self, exc_type, value, traceback):
        if self.sysPathInserted:
            sys.path.remove(self.sysPath)
            self.sysPathInserted = False

def runCode(fileToRun, globals, args, useUntypy=True):
    localDir = os.path.dirname(fileToRun)

    with RunSetup(localDir):
        codeTxt = readFile(fileToRun)
        flags = 0 | anns.compiler_flag
        if useUntypy:
            verbose(f'finding modules imported by {fileToRun}')
            importedMods = findImportedModules([localDir], fileToRun)
            verbose('finished finding modules, now installing import hook on ' + repr(importedMods))
            untypy.just_install_hook(importedMods + ['__wypp__'])
            verbose(f"transforming {fileToRun} for typechecking")
            tree = compile(codeTxt, fileToRun, 'exec', flags=(flags | ast.PyCF_ONLY_AST),
                            dont_inherit=True, optimize=-1)
            untypy.transform_tree(tree, fileToRun)
            verbose(f'done with transformation of {fileToRun}')
            code = tree
        else:
            code = codeTxt
        compiledCode = compile(code, fileToRun, 'exec', flags=flags, dont_inherit=True)
        oldArgs = sys.argv
        try:
            sys.argv = [fileToRun] + args
            exec(compiledCode, globals)
        finally:
            sys.argv = oldArgs

def runStudentCode(fileToRun, globals, onlyCheckRunnable, args, useUntypy=True):
    doRun = lambda: runCode(fileToRun, globals, args, useUntypy=useUntypy)
    if onlyCheckRunnable:
        try:
            doRun()
        except:
            printStderr('Loading file %s crashed' % fileToRun)
            handleCurrentException()
        else:
            die(0)
    doRun()

# globals already contain libDefs
def runTestsInFile(testFile, globals, libDefs, useUntypy=True):
    printStderr()
    printStderr(f"Running tutor's tests in {testFile}")
    libDefs.resetTestCount()
    try:
        runCode(testFile, globals, [], useUntypy=useUntypy)
    except:
        handleCurrentException()
    return libDefs.dict['printTestResults']('Tutor:  ')

# globals already contain libDefs
def performChecks(check, testFile, globals, libDefs, useUntypy=True):
    prefix = ''
    if check and testFile:
        prefix = 'Student: '
    testResultsStudent = libDefs.printTestResults(prefix)
    if check:
        testResultsInstr = {'total': 0, 'failing': 0}
        if testFile:
            testResultsInstr = runTestsInFile(testFile, globals, libDefs, useUntypy=useUntypy)
        failingSum = testResultsStudent['failing'] + testResultsInstr['failing']
        die(0 if failingSum < 1 else 1)

def prepareInteractive(reset=True):
    print()
    if reset:
        if os.name == 'nt':
            # clear the terminal
            os.system('cls')
        else:
            # On linux & mac use ANSI Sequence for this
            print('\033[2J\033[H')

def enterInteractive(userDefs):
    for k, v in userDefs.items():
        globals()[k] = v
    print()

def tbToFrameList(tb):
    cur = tb
    res = []
    while cur:
        res.append(cur.tb_frame)
        cur = cur.tb_next
    return res

def ignoreFrame(frame):
    if DEBUG:
        return False
    modName = frame.f_globals["__name__"]
    return '__wypp_runYourProgram' in frame.f_globals or \
        modName == 'untypy' or modName.startswith('untypy.') or \
        modName == 'wypp' or modName.startswith('wypp.')

# Returns a StackSummary object
def limitTraceback(tb):
    frames = [(f, f.f_lineno) for f in tbToFrameList(tb) if not ignoreFrame(f)]
    return traceback.StackSummary.extract(frames)

def handleCurrentException(exit=True, removeFirstTb=False, file=sys.stderr):
    (etype, val, tb) = sys.exc_info()
    if isinstance(val, SystemExit):
        die(val.code)
    if tb and removeFirstTb:
        tb = tb.tb_next
    stackSummary = limitTraceback(tb)
    header = False
    for x in stackSummary.format():
        if not header:
            file.write('Traceback (most recent call last):\n')
            header = True
        file.write(x)
    if isinstance(val, untypy.error.UntypyError):
        name = 'Wypp' + val.simpleName()
        file.write(name)
        s = str(val)
        if s and s[0] != '\n':
            file.write(': ')
        file.write(s)
        file.write('\n')
    else:
        for x in traceback.format_exception_only(etype, val):
            file.write(x)
    if exit:
        die(1)

HISTORY_SIZE = 1000

def getHistoryFilePath():
    envVar = 'HOME'
    if os.name == 'nt':
        envVar = 'USERPROFILE'
    d = os.getenv(envVar, None)
    if d:
        return os.path.join(d, ".wypp_history")
    else:
        return None

# We cannot import untypy at the top of the file because we might have to install it first.
def importUntypy():
    global untypy
    try:
        import untypy
    except ModuleNotFoundError as e:
        printStderr(f"Module untypy not found, sys.path={sys.path}: {e}")
        die(1)

def versionOk(v):
    if v.major < 3 or v.minor < 9:
        return False
    if v.major == 3 and v.minor == 9 and v.micro < 2:
        return False
    else:
        return True

def main(globals, argList=None):
    v = sys.version_info
    if not versionOk(v):
        vStr = sys.version.split()[0]
        print(f"""
Python in version 3.9.2 or newer is required. You are still using version {vStr}, please upgrade!
""")
        sys.exit(1)

    (args, restArgs) = parseCmdlineArgs(argList)
    global VERBOSE
    if args.verbose:
        VERBOSE = True
    global DEBUG
    if args.debug:
        DEBUG = True

    verbose(f'VERBOSE={VERBOSE}, DEBUG={DEBUG}')

    installLib(args.installMode)
    if site.USER_SITE not in sys.path:
        if not site.ENABLE_USER_SITE:
            printStderr(f"User site-packages disabled ({site.USER_SITE}. This might cause problems importing wypp or untypy.")
        else:
            verbose(f"Adding user site-package directory {site.USER_SITE} to sys.path")
            sys.path.append(site.USER_SITE)
    importUntypy()

    fileToRun = args.file
    if args.changeDir:
        os.chdir(os.path.dirname(fileToRun))
        fileToRun = os.path.basename(fileToRun)

    isInteractive = args.interactive
    version = readVersion()
    if isInteractive:
        prepareInteractive(reset=not args.noClear)

    if fileToRun is None:
        return
    if not args.checkRunnable and (not args.quiet or args.verbose):
        printWelcomeString(fileToRun, version, useUntypy=args.checkTypes)

    libDefs = prepareLib(onlyCheckRunnable=args.checkRunnable)

    globals['__name__'] = '__wypp__'
    sys.modules['__wypp__'] = sys.modules['__main__']
    try:
        verbose(f'running code in {fileToRun}')
        globals['__file__'] = fileToRun
        runStudentCode(fileToRun, globals, args.checkRunnable, restArgs,
                       useUntypy=args.checkTypes)
    except Exception as e:
        verbose(e)
        handleCurrentException()

    performChecks(args.check, args.testFile, globals, libDefs, useUntypy=args.checkTypes)

    if isInteractive:
        enterInteractive(globals)
        if args.checkTypes:
            consoleClass = TypecheckedInteractiveConsole
        else:
            consoleClass = code.InteractiveConsole
        historyFile = getHistoryFilePath()
        try:
            import readline
            readline.parse_and_bind('tab: complete')
            if historyFile and os.path.exists(historyFile):
                readline.read_history_file(historyFile)
        except:
            pass
        try:
            consoleClass(locals=globals).interact(banner="", exitmsg='')
        finally:
            if readline and historyFile:
                readline.set_history_length(HISTORY_SIZE)
                readline.write_history_file(historyFile)

class TypecheckedInteractiveConsole(code.InteractiveConsole):

    def showtraceback(self) -> None:
        handleCurrentException(exit=False, removeFirstTb=True, file=sys.stdout)

    def runsource(self, source, filename="<input>", symbol="single"):
        try:
            code = self.compile(source, filename, symbol)
        except (OverflowError, SyntaxError, ValueError):
            self.showsyntaxerror(filename)
            return False
        if code is None:
            return True
        try:
            import ast
            tree = compile("\n".join(self.buffer), filename, symbol, flags=ast.PyCF_ONLY_AST, dont_inherit=True, optimize=-1)
            untypy.transform_tree(tree, filename)
            code = compile(tree, filename, symbol)
        except Exception as e:
            if hasattr(e, 'text') and e.text == "":
                pass
            else:
                traceback.print_tb(e.__traceback__)
        self.runcode(code)
        return False



from __future__ import annotations

import inspect
from typing import Optional, Any, Callable, TypeVar, List, Tuple

from untypy.error import UntypyTypeError, Location, UntypyAttributeError


class CreationContext:
    def find_checker(self, annotation: Any) -> Optional[TypeChecker]:
        raise NotImplementedError

    def declared_location(self) -> Location:
        raise NotImplementedError

    def wrap(self, err: UntypyAttributeError) -> UntypyAttributeError:
        raise NotImplementedError

    def resolve_typevar(self, var: TypeVar) -> Tuple[bool, Any]:
        raise NotImplementedError

    def all_typevars(self) -> List[TypeVar]:
        raise NotImplementedError

    def with_typevars(self, typevars: dict[TypeVar, Any]) -> CreationContext:
        raise NotImplementedError

    def should_be_inheritance_checked(self, annotation: type) -> bool:
        raise NotImplementedError

    def eval_context(self):
        raise NotImplementedError


class ExecutionContext:
    def wrap(self, err: UntypyTypeError) -> UntypyTypeError:
        raise NotImplementedError


class TypeChecker:

    def describe(self) -> str:
        raise NotImplementedError

    def may_be_wrapped(self) -> bool:
        return False

    def base_type(self) -> list[Any]:
        raise NotImplementedError

    # Higher Priority => checked first inside Union.
    def base_type_priority(self) -> int:
        return 0

    def check_and_wrap(self, arg: Any, ctx: ExecutionContext) -> Any:
        raise NotImplementedError


class TypeCheckerFactory:

    def create_from(self, annotation: Any, ctx: CreationContext) -> Optional[TypeChecker]:
        raise NotImplementedError


WrappedFunctionContextProvider = Callable[[str], ExecutionContext]


class WrappedFunction:
    def get_original(self):
        raise NotImplementedError

    def wrap_arguments(self, ctxprv: WrappedFunctionContextProvider, args, kwargs):
        raise NotImplementedError

    def wrap_return(self, ret, bindings, ctx: ExecutionContext):
        raise NotImplementedError

    def describe(self) -> str:
        raise NotImplementedError

    def checker_for(self, name: str) -> TypeChecker:
        raise NotImplementedError

    @staticmethod
    def find_original(fn):
        if hasattr(fn, '__original'):
            return WrappedFunction.find_original(getattr(fn, '__original'))
        elif isinstance(fn, WrappedFunction):
            return WrappedFunction.find_original(fn.get_original())
        elif hasattr(fn, '__wf'):
            return WrappedFunction.find_original(getattr(fn, '__wf').get_original())
        else:
            return fn

    @staticmethod
    def find_location(fn) -> Optional[Location]:
        fn = WrappedFunction.find_original(fn)
        try:
            return Location(
                file=inspect.getfile(fn),
                line_no=inspect.getsourcelines(fn)[1],
                line_span=len(inspect.getsourcelines(fn)[0]),
            )
        except:  # Failes on builtins
            return None



from typing import Generic, TypeVar, Optional, Tuple, Iterable, Any, Iterator

I = TypeVar("I")


class Set(Generic[I], set):

    def add(self, other: I) -> None:
        pass

    def clear(self) -> None:
        pass

    def discard(self, elem: I):
        pass

    def pop(self) -> Optional[I]:
        pass

    def remove(self, elem: I) -> None:
        pass

    def update(self, *others: Tuple[Iterable[I], ...]) -> None:
        pass

    # This method returns `NotImplemented`, i don't know why.
    def __ior__(self, *others: Tuple[Iterable[I], ...]) -> Any:
        pass

    def __contains__(self, key: I) -> bool:
        pass

    def __iter__(self) -> Iterator[I]:
        pass

    def __len__(self) -> int:
        pass

    # Only removes elements. No checking needed. Argument type is set Any
    #
    # def intersection_update(self, others: Iterator[set[Any]]) -> None:
    #     pass
    #
    # def __iand__(self, other: set[Any]) -> None:
    #     pass
    #
    # def difference_update(self, others: Iterator[set[Any]]) -> None:
    #     pass
    #
    # def __isub__(self, other: set[Any]) -> None:
    #     pass

    # Mutable meth
    # symmetric_difference_update
    # __ixor__

    # Recursion
    # copy

    # Immutable generic meth
    # difference
    # intersection
    # isdisjoint
    # issubset
    # issuperset
    # symmetric_difference(self, other)
    # union(self, other)
    # __and__
    # __or__
    # __rand__
    # __reduce__ ???
    # __ror__
    # __rsub__
    # __rxor__
    # __sub__
    # __xor__



import inspect
import sys
import typing
from typing import Protocol, Any, Optional, Callable, Union, TypeVar, Dict, Tuple
import untypy.util.display as display
from untypy.error import UntypyTypeError, UntypyAttributeError, Frame, Location, ResponsibilityType
from untypy.impl.any import SelfChecker, AnyChecker
from untypy.interfaces import TypeCheckerFactory, CreationContext, TypeChecker, ExecutionContext, \
    WrappedFunctionContextProvider
from untypy.util import WrappedFunction, ArgumentExecutionContext, ReturnExecutionContext
from untypy.util.condition import FunctionCondition
from untypy.util.typehints import get_type_hints


class ProtocolFactory(TypeCheckerFactory):

    def create_from(self, annotation: Any, ctx: CreationContext) -> Optional[TypeChecker]:
        if isinstance(annotation, type) and Protocol in annotation.mro():
            return ProtocolChecker(annotation, ctx)
        elif hasattr(annotation, '__args__') and hasattr(annotation.__origin__,
                                                         '__mro__') and typing.Protocol in annotation.__origin__.__mro__:
            return ProtocolChecker(annotation, ctx)
        else:
            return None


def _find_bound_typevars(clas: type) -> Tuple[type, Dict[TypeVar, Any]]:
    if not hasattr(clas, '__args__') or not hasattr(clas, '__origin__'):
        return (clas, dict())
    if not hasattr(clas.__origin__, '__parameters__'):
        return (clas, dict())

    keys = clas.__origin__.__parameters__
    values = clas.__args__

    if len(keys) != len(values):
        raise UntypyAttributeError(f"Some unbound Parameters in {clas.__name__}. "
                                   f"keys={keys} do not match values={values}.",
                                   [Location(
                                       file=inspect.getfile(clas),
                                       line_no=inspect.getsourcelines(clas)[1],
                                       line_span=len(inspect.getsourcelines(clas)[0]))])
    return (clas.__origin__, dict(zip(keys, values)))


def get_proto_members(proto: type, ctx: CreationContext) -> dict[
    str, Tuple[inspect.Signature, dict[str, TypeChecker], FunctionCondition]]:
    blacklist = ['__init__', '__class__', '__delattr__', '__dict__', '__dir__',
                 '__doc__', '__getattribute__', '__getattr__', '__init_subclass__',
                 '__new__', '__setattr__', '__subclasshook__', '__weakref__',
                 '__abstractmethods__', '__class_getitem__']

    member_dict = {}
    for [name, member] in inspect.getmembers(proto):
        if name in blacklist:
            continue

        if inspect.isfunction(member):
            member = WrappedFunction.find_original(member)
            signature = inspect.signature(member)

            is_typed = len(inspect.getfullargspec(member).annotations) != 0

            checkers = {}
            if not is_typed:
                # Use Any for any type
                for key in signature.parameters:
                    if key == 'self':
                        checkers[key] = SelfChecker()
                    else:
                        checkers[key] = AnyChecker()
                checkers['return'] = AnyChecker()
            else:
                annotations = get_type_hints(member, ctx)
                for key in signature.parameters:
                    if key == 'self':
                        checkers[key] = SelfChecker()
                    else:
                        param = signature.parameters[key]
                        if param.annotation is inspect.Parameter.empty:
                            raise ctx.wrap(UntypyAttributeError(
                                f"Missing annotation for argument '{key}' of function {member.__name__} "
                                f"in protocol {proto.__name__}\n"))

                        param_anot = annotations[key]
                        if param_anot is proto:
                            checker = SimpleInstanceOfChecker(proto, None)
                        else:
                            checker = ctx.find_checker(param_anot)
                        if checker is None:
                            raise ctx.wrap(UntypyAttributeError(f"\n\tUnsupported type annotation: {param.annotation}\n"
                                                                f"for argument '{key}' of function {member.__name__} "
                                                                f"in protocol {proto.__name__}.\n"))
                        checkers[key] = checker

                if signature.return_annotation is inspect.Parameter.empty:
                    return_annotation = None
                else:
                    return_annotation = annotations['return']
                if return_annotation is proto:  # Self as Return Type would led to endless recursion
                    return_checker = SimpleInstanceOfChecker(proto, None)
                else:
                    return_checker = ctx.find_checker(return_annotation)

                if return_checker is None:
                    raise ctx.wrap(UntypyAttributeError(f"\n\tUnsupported type annotation: {signature.return_annotation}\n"
                                                        f"for return value of function {member.__name__} "
                                                        f"in protocol-like {proto.__name__}.\n"))
                checkers['return'] = return_checker

            fc = None
            if hasattr(member, '__fc'):
                fc = getattr(member, '__fc')
            member_dict[name] = (signature, checkers, fc)
    return member_dict


class ProtocolChecker(TypeChecker):
    def __init__(self, annotation: type, ctx: CreationContext, *, altname : Optional[str] = None,
            omit_tyargs=False):
        (proto, typevars) = _find_bound_typevars(annotation)
        self.ctx = ctx.with_typevars(typevars)
        self.proto = proto
        self._members = None
        self.typevars = typevars
        self.wrapper_types = dict()
        self.altname = altname
        self.omit_tyargs = omit_tyargs

    @property
    def members(self):
        if not self._members:
            self._members = get_proto_members(self.proto, self.ctx)
        return self._members

    def may_change_identity(self) -> bool:
        return True

    def check_and_wrap(self, arg: Any, ctx: ExecutionContext) -> Any:
        if hasattr(arg, '_ProtocolWrappedFunction__inner'):
            # no double wrapping
            arg = getattr(arg, '_ProtocolWrappedFunction__inner')

        if type(arg) in self.wrapper_types:
            wrapped_type = self.wrapper_types[type(arg)]
        else:
            wrapped_type = ProtocolWrapper(self, arg, self.members, ctx)
            self.wrapper_types[type(arg)] = wrapped_type
        # On wrappers some built-in classes like tuple, the constructor can not
        # be called directly, because it will trigger the original one.
        # To me this looks like a bug in the interpreter.
        # return wrapped_type(arg, ctx)
        w = wrapped_type.__new__(wrapped_type)
        w.__init__(arg, ctx)
        return w


    def base_type(self) -> list[Any]:
        # Prevent Classes implementing multiple Protocols in one Union by accident.
        return [Protocol]

    def describe(self) -> str:
        if self.altname is not None:
            return self.altname

        desc = set([])
        if not self.omit_tyargs:
            for name in self.members:
                (sig, binds, cond) = self.members[name]
                for argname in sig.parameters:
                    if isinstance(sig.parameters[argname].annotation, TypeVar):
                        desc.add(binds[argname].describe())
                if isinstance(sig.return_annotation, TypeVar):
                    desc.add(binds['return'].describe())
        if len(desc) > 0:
            # FIXME: what about the ordering of tyvars?
            return f"{self.proto.__name__}[" + (', '.join(desc)) + "]"
        else:
            return f"{self.proto.__name__}"

    def protocol_type(self) -> str:
        return f"protocol"

    def protoname(self):
        return self.describe()

def isInternalProtocol(p: Any):
    if isinstance(p, ProtocolChecker):
        p = p.proto
    if hasattr(p, '__module__'):
        return 'untypy.' in p.__module__
    else:
        return False

def protoMismatchErrorMessage(what: str, proto: Any):
    if isinstance(proto, ProtocolChecker):
        kind = proto.protocol_type()
        name = proto.protoname()
    else:
        kind = 'protocol'
        name = proto.__name__
    isUserDefined = True
    if isInternalProtocol(proto):
        isUserDefined = False
    if isUserDefined:
        return f"{what} does not implement {kind} {name}"
    else:
        return f"{what} is not {display.withIndefiniteArticle(name)}"

def ProtocolWrapper(protocolchecker: ProtocolChecker, originalValue: Any,
                    members: dict[str, Tuple[inspect.Signature, dict[str, TypeChecker], FunctionCondition]],
                    ctx: ExecutionContext):
    list_of_attr = dict()
    original = type(originalValue)
    for fnname in members:
        if not hasattr(original, fnname):
            err = ctx.wrap(UntypyTypeError(
                expected=protocolchecker.describe(),
                given=originalValue
            )).with_header(
                protoMismatchErrorMessage(original.__name__, protocolchecker.proto)
            )
            missing = []
            for fnname in members:
                if not hasattr(original, fnname):
                    missing.append(fnname)
            if len(missing) == 2:
                err = err.with_note(f"It is missing the functions '{missing[0]}' and '{missing[1]}'")
            elif len(missing) == 1:
                err = err.with_note(f"It is missing the function '{missing[0]}'")
            raise err

        original_fn = getattr(original, fnname)
        try:
            # fails on built ins - YEAH
            original_fn_signature = inspect.signature(original_fn)
        except:
            original_fn_signature = None

        if hasattr(original_fn, '__wf'):
            original_fn = getattr(original_fn, '__wf')
        (sig, argdict, fc) = members[fnname]

        for param in sig.parameters:
            if original_fn_signature is not None and param not in original_fn_signature.parameters:
                raise ctx.wrap(UntypyTypeError(
                    expected=protocolchecker.describe(),
                    given=originalValue
                )).with_header(
                    protoMismatchErrorMessage(original.__name__, protocolchecker.proto)
                ).with_note(f"The signature of '{fnname}' does not match. Missing required parameter {param}.")

        list_of_attr[fnname] = ProtocolWrappedFunction(original_fn, sig, argdict, protocolchecker, fc).build()

    def constructor(me, inner, ctx):
        me._ProtocolWrappedFunction__inner = inner
        me._ProtocolWrappedFunction__ctx = ctx

    def __repr__(me):
        return me._ProtocolWrappedFunction__inner.__repr__()

    def __str__(me):
        return me._ProtocolWrappedFunction__inner.__str__()

    def __getattr__(me, name):
        return getattr(me._ProtocolWrappedFunction__inner, name)

    def __setattr__(me, name, value):
        if name == '_ProtocolWrappedFunction__inner':
            super(type(me), me).__setattr__('_ProtocolWrappedFunction__inner', value)
            return
        if name == '_ProtocolWrappedFunction__ctx':
            super(type(me), me).__setattr__('_ProtocolWrappedFunction__ctx', value)
            return

        return setattr(me._ProtocolWrappedFunction__inner, name, value)

    list_of_attr['__init__'] = constructor
    list_of_attr['__getattr__'] = __getattr__  # allow access of attributes
    list_of_attr['__setattr__'] = __setattr__  # allow access of attributes
    list_of_attr['__repr__'] = __repr__
    list_of_attr['__str__'] = __str__

    name = f"WyppTypeCheck({original.__name__})"

    if type(original) == type and original.__flags__ & 0x0400 and original not in [dict, list, set, tuple, str]:
        # This class does not have any metaclass that may have unexpected side effects.
        # Also the Py_TPFLAGS_BASETYPE=0x0400 must be set to inheritable, as some classes like C-Based classes
        # like`dict_items` can not be inherited from.
        # Also some other built-in types have bugs when inherited from.
        orig_tuple = (original,)
    else:
        # Fall back to no inheritance, this should be an edge case.
        orig_tuple = ()

    t = type(name, orig_tuple, list_of_attr)

    if hasattr(original, '__module__'):
        t.__module__ = original.__module__

    return t

class ProtocolWrappedFunction(WrappedFunction):

    def __init__(self, inner: Union[Callable, WrappedFunction], signature: inspect.Signature,
                 checker: Dict[str, TypeChecker],
                 protocol: ProtocolChecker,
                 fc: FunctionCondition):
        self.inner = inner
        self.signature = signature
        self.checker = checker
        self.protocol = protocol
        self.fc = fc

    def build(self):
        fn = WrappedFunction.find_original(self.inner)

        fn_of_protocol = getattr(self.protocol.proto, fn.__name__)
        if hasattr(fn_of_protocol, '__wf'):
            fn_of_protocol = getattr(fn_of_protocol, '__wf')

        def wrapper(me, *args, **kwargs):
            inner_object = me.__inner
            inner_ctx = me.__ctx

            caller = sys._getframe(1)
            (args, kwargs, bind1) = self.wrap_arguments(lambda n: ArgumentExecutionContext(fn_of_protocol, caller, n),
                                                        (inner_object, *args), kwargs)
            if isinstance(self.inner, WrappedFunction):
                (args, kwargs, bind2) = self.inner.wrap_arguments(lambda n:
                                                                  ProtocolArgumentExecutionContext(self, n,
                                                                                                   inner_object,
                                                                                                   inner_ctx),
                                                                  args, kwargs)
            ret = fn(*args, **kwargs)
            if isinstance(self.inner, WrappedFunction):
                ret = self.inner.wrap_return(ret, bind2, ProtocolReturnExecutionContext(self,
                                                                                        ResponsibilityType.IN,
                                                                                        inner_object,
                                                                                        inner_ctx))
            return self.wrap_return(ret, bind1, ProtocolReturnExecutionContext(self,
                                                                               ResponsibilityType.OUT, inner_object,
                                                                               inner_ctx))

        async def async_wrapper(*args, **kwargs):
            raise AssertionError("Not correctly implemented see wrapper")

        if inspect.iscoroutine(self.inner):
            w = async_wrapper
        else:
            w = wrapper

        setattr(w, '__wrapped__', fn)
        setattr(w, '__name__', fn.__name__)
        setattr(w, '__signature__', self.signature)
        setattr(w, '__wf', self)
        return w

    def get_original(self):
        return self.inner

    def wrap_arguments(self, ctxprv: WrappedFunctionContextProvider, args, kwargs):
        try:
            bindings = self.signature.bind(*args, **kwargs)
        except TypeError as e:
            err = UntypyTypeError(header=str(e))
            if "self" not in self.signature.parameters:
                err = err.with_note("Hint: 'self'-parameter was omitted in declaration.")
            raise ctxprv("").wrap(err)

        bindings.apply_defaults()
        if self.fc is not None:
            self.fc.prehook(bindings, ctxprv)
        for name in bindings.arguments:
            check = self.checker[name]
            ctx = ctxprv(name)
            bindings.arguments[name] = check.check_and_wrap(bindings.arguments[name], ctx)
        return bindings.args, bindings.kwargs, bindings

    def wrap_return(self, ret, bindings, ctx: ExecutionContext):
        check = self.checker['return']
        if self.fc is not None:
            self.fc.posthook(ret, bindings, ctx)
        return check.check_and_wrap(ret, ctx)

    def describe(self) -> str:
        fn = WrappedFunction.find_original(self.inner)
        return f"{fn.__name__}" + str(self.signature)

    def checker_for(self, name: str) -> TypeChecker:
        return self.checker[name]

    def declared(self) -> Location:
        fn = WrappedFunction.find_original(self.inner)
        return WrappedFunction.find_location(getattr(self.protocol.proto, fn.__name__))


class ProtocolReturnExecutionContext(ExecutionContext):
    def __init__(self, wf: ProtocolWrappedFunction, invert: ResponsibilityType, me: Any, ctx: ExecutionContext):
        self.wf = wf
        self.invert = invert
        self.me = me
        self.ctx = ctx

    def wrap(self, err: UntypyTypeError) -> UntypyTypeError:
        err = ReturnExecutionContext(self.wf).wrap(err)

        if err.responsibility_type is self.invert:
            return err
        responsable = WrappedFunction.find_location(self.wf)
        (decl, ind) = err.next_type_and_indicator()
        err = err.with_inverted_responsibility_type()
        err = err.with_frame(Frame(
            decl,
            ind,
            declared=self.wf.declared(),
            responsable=responsable
        ))

        inner = self.wf.inner
        if isinstance(inner, WrappedFunction):
            err = err.with_note(
                f"The return value of method '{WrappedFunction.find_original(self.wf).__name__}' does violate the {self.wf.protocol.protocol_type()} '{self.wf.protocol.proto.__name__}'.")
            err = err.with_note(
                f"The annotation '{inner.checker_for('return').describe()}' is incompatible with the {self.wf.protocol.protocol_type()}'s annotation '{self.wf.checker_for('return').describe()}'\nwhen checking against the following value:")

        previous_chain = UntypyTypeError(
            self.me,
            f"{self.wf.protocol.protoname()}"
        ).with_header(
            protoMismatchErrorMessage(type(self.me).__name__, self.wf.protocol)
        )

        previous_chain = self.ctx.wrap(previous_chain)
        if isInternalProtocol(self.wf.protocol):
            return previous_chain
        else:
            return err.with_previous_chain(previous_chain)



class ProtocolArgumentExecutionContext(ExecutionContext):
    def __init__(self, wf: ProtocolWrappedFunction, arg_name: str, me: Any, ctx: ExecutionContext):
        self.wf = wf
        self.arg_name = arg_name
        self.me = me
        self.ctx = ctx

    def wrap(self, err: UntypyTypeError) -> UntypyTypeError:
        (original_expected, _ind) = err.next_type_and_indicator()
        err = ArgumentExecutionContext(self.wf, None, self.arg_name).wrap(err)

        responsable = WrappedFunction.find_location(self.wf)

        (decl, ind) = err.next_type_and_indicator()
        err = err.with_frame(Frame(
            decl,
            ind,
            declared=self.wf.declared(),
            responsable=responsable
        ))

        err = err.with_note(
            f"Argument {self.arg_name} of method {WrappedFunction.find_original(self.wf).__name__} violates the type declared by the {self.wf.protocol.protocol_type()} {self.wf.protocol.proto.__name__}.")
        err = err.with_note(
            f"Annotation {original_expected} is incompatible with the {self.wf.protocol.protocol_type()}'s annotation {self.wf.checker_for(self.arg_name).describe()}.")

        previous_chain = UntypyTypeError(
            self.me,
            f"{self.wf.protocol.protoname()}"
        ).with_header(
            protoMismatchErrorMessage(type(self.me).__name__, self.wf.protocol)
        )
        previous_chain = self.ctx.wrap(previous_chain)
        if isInternalProtocol(self.wf.protocol):
            return previous_chain
        else:
            return err.with_previous_chain(previous_chain)


class SimpleInstanceOfChecker(TypeChecker):
    def __init__(self, annotation: type, ctx: CreationContext):
        self.annotation = annotation

    def check_and_wrap(self, arg: Any, ctx: ExecutionContext) -> Any:
        if isinstance(arg, self.annotation):
            return arg
        else:
            raise ctx.wrap(UntypyTypeError(arg, self.describe()))

    def describe(self) -> str:
        return self.annotation.__name__

    def base_type(self) -> Any:
        return [self.annotation]
